"""
Contains relevant physical constants
"""

import collections
from decimal import Decimal
from functools import lru_cache
from typing import TYPE_CHECKING, Union

from ..datum import Datum, print_variables
from .ureg import build_units_registry

if TYPE_CHECKING:
    from pint import quantity, UnitRegistry  # lgtm: [py/unused-import]


class PhysicalConstantsContext:
    """CODATA physical constants set from NIST.

    Parameters
    ----------
    context : {'CODATA2014', 'CODATA2018'}
        Origin of loaded data.

    Attributes
    ----------
    doi : str
        The DOI of the current context.
    name : str
        The name of the context ('CODATA2014')
    pc : dict of Datum
        Each physical constant is an entry in `pc`, where key is the
        lowercased string of the NIST name (or any alias) and the
        value is a Datum object with `lbl` the exact NIST name string,
        `units`, `data` value as Decimal object, and any uncertainty
        in the `comment` field.
    raw_codata : Dict[str, Any]
        A dictionary representation of the raw context data.
    year : int
        The year the context was created.
    """

    _transtable = str.maketrans(" -/{", "__p_", ".,()}")

    # Alias Typing
    h: float
    c: float
    kb: float
    R: float
    bohr2angstroms: float
    bohr2cm: float
    amu2g: float
    amu2kg: float
    hartree2J: float
    hartree2aJ: float
    cal2J: float
    dipmom_au2si: float
    dipmom_au2debye: float
    c_au: float
    hartree2ev: float
    hartree2wavenumbers: float
    hartree2kcalmol: float
    hartree2kJmol: float
    hartree2MHz: float
    kcalmol2wavenumbers: float
    e0: float
    na: float
    me: float

    def __init__(self, context="CODATA2014"):
        self.pc = collections.OrderedDict()

        if context == "CODATA2014":
            from ..data import nist_2014_codata

            self.doi = nist_2014_codata["doi"]
            self.raw_codata = nist_2014_codata["constants"]
        elif context == "CODATA2018":
            from ..data import nist_2018_codata

            self.doi = nist_2018_codata["doi"]
            self.raw_codata = nist_2018_codata["constants"]

        else:
            raise KeyError("Context set as '{}', only contexts {'CODATA2014', 'CODATA2018', } are currently supported")

        # physical constant loop
        for k, v in self.raw_codata.items():
            self.pc[k] = Datum(
                v["quantity"],
                v["unit"],
                Decimal(v["value"]),
                comment="uncertainty={}".format(v["uncertainty"]),
                doi=self.doi,
            )

        self.name = context
        self.year = int(context.replace("CODATA", ""))
        self._ureg = None

        # Extra relationships
        self.pc["calorie-joule relationship"] = Datum(
            "calorie-joule relationship", "J", Decimal("4.184"), comment="uncertainty=(exact)"
        )

        rename_2018_from_2014 = {
            "atomic unit of momentum": "atomic unit of mom.um",
            "reduced Planck constant": "Planck constant over 2 pi",
            "reduced Planck constant in eV s": "Planck constant over 2 pi in eV s",
            "reduced Planck constant times c in MeV fm": "Planck constant over 2 pi times c in MeV fm",
            "natural unit of momentum": "natural unit of mom.um",
            "natural unit of momentum in MeV/c": "natural unit of mom.um in MeV/c",
            "electron gyromag. ratio in MHz/T": "electron gyromag. ratio over 2 pi",
            "vacuum mag. permeability": "mag. constant",
            "lattice spacing of ideal Si (220)": "{220} lattice spacing of silicon",
            "Planck constant in eV/Hz": "Planck constant in eV s",
            "Bohr magneton in inverse meter per tesla": "Bohr magneton in inverse meters per tesla",
            "Boltzmann constant in inverse meter per kelvin": "Boltzmann constant in inverse meters per kelvin",
            "Copper x unit": "Cu x unit",
            "Molybdenum x unit": "Mo x unit",
            "proton gyromag. ratio in MHz/T": "proton gyromag. ratio over 2 pi",
            "shielded proton gyromag. ratio in MHz/T": "shielded proton gyromag. ratio over 2 pi",
            "reduced proton Compton wavelength": "proton Compton wavelength over 2 pi",
            "reduced tau Compton wavelength": "tau Compton wavelength over 2 pi",
            "tau energy equivalent": "tau mass energy equivalent in mev",
            "reduced neutron Compton wavelength": "neutron Compton wavelength over 2 pi",
            "neutron gyromag. ratio in MHz/T": "neutron gyromag. ratio over 2 pi",
            "nuclear magneton in inverse meter per tesla": "nuclear magneton in inverse meters per tesla",
            "shielded helion gyromag. ratio in MHz/T": "shielded helion gyromag. ratio over 2 pi",
            "reduced Compton wavelength": "Compton wavelength over 2 pi",
            "vacuum electric permittivity": "electric constant",
            "reduced muon Compton wavelength": "muon Compton wavelength over 2 pi",
        }

        # fmt: off
        if context == "CODATA2014":
            aliases = []

        elif context == "CODATA2018":
            for new_name, old_name in rename_2018_from_2014.items():
                dm = self.pc[new_name.lower()]
                self.pc[old_name.lower()] = Datum(old_name, dm.units, dm.data, comment=dm.comment, doi=dm.doi)

            aliases = [
                ("molar Planck constant times c",                      "J m mol^{-1}", self.pc["molar planck constant"].data * self.pc["speed of light in vacuum"].data,     ""),
                ("Faraday constant for conventional electric current", "C_{90} mol^{-1}", self.pc["faraday constant"].data / self.pc["conventional value of coulomb-90"].data, ""),
                ("elementary charge over h",                           "A J^{-1}", self.pc["elementary charge over h-bar"].data / (2 * _get_pi(from_scratch=False)),    ""),
            ]

        aliases.extend([
            ('h',                    'J s',            self.pc['hertz-joule relationship'].data,                             'The Planck constant (Js)'),
            ('hbar',                 'J s',            self.pc['planck constant over 2 pi'].data,                            'Reduced Planck constant (Js)'),
            ('c',                    'm s^-1',         self.pc['inverse meter-hertz relationship'].data,                     'Speed of light (ms$^{-1}$)'),
            ('kb',                   'J K^-1',         self.pc['kelvin-joule relationship'].data,                            'The Boltzmann constant (JK$^{-1}$)'),
            ('R',                    'J mol^-1 K^-1',  self.pc['molar gas constant'].data,                                   'Universal gas constant (JK$^{-1}$mol$^{-1}$)'),
            ('bohr2angstroms',       'AA',             self.pc['bohr radius'].data * Decimal('1.E10'),                       'Bohr to Angstroms conversion factor'),
            ('bohr2m',               'm',              self.pc['bohr radius'].data,                                          'Bohr to meters conversion factor'),
            ('bohr2cm',              'cm',             self.pc['bohr radius'].data * Decimal('100'),                         'Bohr to centimeters conversion factor'),
            ('amu2g',                'g',              self.pc['atomic mass constant'].data * Decimal('1000'),               'Atomic mass units to grams conversion factor'),
            ('amu2kg',               'kg',             self.pc['atomic mass constant'].data,                                 'Atomic mass units to kg conversion factor' ),
            ('au2amu',               'u',              self.pc['electron mass in u'].data,                                   'Atomic units (m$@@e$) to atomic mass units conversion factor'),
            ('hartree2J',            'J',              self.pc['hartree energy'].data,                                       'Hartree to joule conversion factor'),
            ('hartree2aJ',           'aJ',             self.pc['hartree energy'].data * Decimal('1.E18'),                    'Hartree to attojoule (10$^{-18}$J) conversion factor'),
            ('cal2J',                'J',              self.pc['calorie-joule relationship'].data,                           'Calorie to joule conversion factor'),
            ('dipmom_au2si',         'C m',            self.pc['atomic unit of electric dipole mom.'].data,                  'Atomic units to SI units (Cm) conversion factor for dipoles'),
            ('dipmom_au2debye',      '???',            self.pc['atomic unit of electric dipole mom.'].data * Decimal('1.E21') / self.pc['hertz-inverse meter relationship'].data,
                                                                                                                             'Atomic units to Debye conversion factor for dipoles'),
            ('dipmom_debye2si',      'C m',            self.pc['hertz-inverse meter relationship'].data * Decimal('1.E-21'), 'Debye to SI units (Cm) conversion factor for dipoles'),
            ('c_au',                 '',               self.pc['inverse fine-structure constant'].data,                      'Speed of light in atomic units'),
            ('hartree2ev',           'eV',             self.pc['hartree energy in ev'].data,                                 'Hartree to eV conversion factor'),
            ('hartree2wavenumbers',  'cm^-1',          self.pc['hartree-inverse meter relationship'].data * Decimal('0.01'), 'Hartree to cm$^{-1}$ conversion factor'),
            ('hartree2kcalmol',      'kcal mol^-1',    self.pc['hartree energy'].data * self.pc['avogadro constant'].data * Decimal('0.001') / self.pc['calorie-joule relationship'].data,
                                                                                                                             'Hartree to kcal mol$^{-1}$ conversion factor'),
            ('hartree2kJmol',        'kJ mol^-1',      self.pc['hartree energy'].data * self.pc['avogadro constant'].data * Decimal('0.001'), 'Hartree to kilojoule mol$^{-1}$ conversion factor'),
            ('hartree2MHz',          'MHz',            self.pc['hartree-hertz relationship'].data * Decimal('1.E-6'),        'Hartree to MHz conversion factor'),
            ('na',                   'mol^-1',         self.pc['avogadro constant'].data,                                    "Avogadro's number"),
            ('me',                   'kg',             self.pc['electron mass'].data,                                        'Electron rest mass (in kg)'),
        ])

        if context == "CODATA2014":
            aliases.extend([
            ('kcalmol2wavenumbers',  'kcal cm mol^-1', Decimal('10') * self.pc['calorie-joule relationship'].data / self.pc['molar planck constant times c'].data,
                                                                                                                             'kcal mol$^{-1}$ to cm$^{-1}$ conversion factor',),
            ('e0',                   'F m^-1',         self.pc['electric constant'].data,                                    'Vacuum permittivity (Fm$^{-1}$)'),
            ])
        elif context == "CODATA2018":
            aliases.extend([
            ("kcalmol2wavenumbers",  "kcal cm mol^-1", Decimal("10") * self.pc["calorie-joule relationship"].data / (self.pc["molar planck constant"].data *
                                                       self.pc["speed of light in vacuum"].data),                            "kcal mol$^{-1}$ to cm$^{-1}$ conversion factor",),
            ("e0",                   "F m^-1",         self.pc["vacuum electric permittivity"].data,                         "Vacuum permittivity (Fm$^{-1}$)"),
            ])
        # fmt: on

        # add alternate names for constants or derived values to help QC programs
        for alias in aliases:
            ident, units, value, comment = alias
            self.pc[ident.lower()] = Datum(ident, units, value, comment=comment)

        # add constants as directly callable member data
        for qca in self.pc.values():
            callname = qca.label.translate(self._transtable)
            setattr(self, callname, float(qca.data))

    def __str__(self) -> str:
        return "PhysicalConstantsContext(context='{}')".format(self.name)

    @property
    def ureg(self) -> "UnitRegistry":
        """Returns the internal Pint units registry.

        Returns
        -------
        UnitRegistry
            The pint context
        """
        if self._ureg is None:
            self._ureg = build_units_registry(self)

        return self._ureg

    def get(self, physical_constant: str, return_tuple: bool = False) -> Union[float, Datum]:
        """Access a physical constant, `physical_constant`.

        Parameters
        ----------
        physical_constant : str
            Case-insensitive string of physical constant with NIST name.
        return_tuple : bool, optional
            See below.

        Returns
        -------
        Union[float, 'Datum']
            When ``return_tuple=False``, value of physical constant.
            When ``return_tuple=True``, Datum with units, description, uncertainty, and value of physical constant as Decimal.

        """
        qca = self.pc[physical_constant.lower()]

        if return_tuple:
            return qca
        else:
            return float(qca.data)

    #       h                         'hertz-joule relationship'                  = 6.62606896E-34       # The Planck constant (Js)
    #       c                         'inverse meter-hertz relationship'          = 2.99792458E8         # Speed of light (ms$^{-1}$)
    #       kb                        'kelvin-joule relationship'                 = 1.3806504E-23        # The Boltzmann constant (JK$^{-1}$)
    #       R                         'molar gas constant'                        = 8.314472             # Universal gas constant (JK$^{-1}$mol$^{-1}$)
    #       bohr2angstroms            'Bohr radius' * 1.E10                       = 0.52917720859        # Bohr to Angstroms conversion factor
    #       bohr2m                    'Bohr radius'                               = 0.52917720859E-10    # Bohr to meters conversion factor
    #       bohr2cm                   'Bohr radius' * 100                         = 0.52917720859E-8     # Bohr to centimeters conversion factor
    #       amu2kg                    'atomic mass constant'                      = 1.660538782E-27      # Atomic mass units to kg conversion factor
    #       au2amu                    'electron mass in u'                        = 5.485799097E-4       # Atomic units (m$@@e$) to atomic mass units conversion factor
    #       hartree2J                 'Hartree energy'                            = 4.359744E-18         # Hartree to joule conversion factor
    #       hartree2aJ                'Hartree energy' * 1.E18                    = 4.359744             # Hartree to attojoule (10$^{-18}$J) conversion factor
    #       cal2J                     = 4.184                # Calorie to joule conversion factor
    #       dipmom_au2si              'atomic unit of electric dipole mom.'       = 8.47835281E-30       # Atomic units to SI units (Cm) conversion factor for dipoles
    #       dipmom_au2debye           'atomic unit of electric dipole mom.' / ('hertz-inverse meter relationship' * 1.E-21)     = 2.54174623           # Atomic units to Debye conversion factor for dipoles
    #       dipmom_debye2si           'hertz-inverse meter relationship' * 1.E-21 = 3.335640952E-30      # Debye to SI units (Cm) conversion factor for dipoles
    #       c_au                      'inverse fine-structure constant'           = 137.035999679        # Speed of light in atomic units
    #       hartree2ev                'Hartree energy in eV'                      = 27.21138             # Hartree to eV conversion factor
    #       hartree2wavenumbers       'hartree-inverse meter relationship' * 0.01 = 219474.6             # Hartree to cm$^{-1}$ conversion factor
    #       hartree2kcalmol           hartree2kJmol / cal2J                       = 627.5095             # Hartree to kcal mol$^{-1}$ conversion factor
    #       hartree2kJmol             'Hartree energy'*'Avogadro constant'*0.001  = 2625.500             # Hartree to kilojoule mol$^{-1}$ conversion factor
    #       hartree2MHz               'hartree-hertz relationship'                = 6.579684E9           # Hartree to MHz conversion factor
    #       kcalmol2wavenumbers       10. / 'molar Planck constant times c'*4.184 = 349.7551             # kcal mol$^{-1}$ to cm$^{-1}$ conversion factor
    #       e0                        'electric constant'                         = 8.854187817E-12      # Vacuum permittivity (Fm$^{-1}$)
    #       na                        'Avogadro constant'                         = 6.02214179E23        # Avogadro's number
    #       me                        'electron mass'                             = 9.10938215E-31       # Electron rest mass (in kg)

    def Quantity(self, data: str) -> "quantity._Quantity":
        """Returns a Pint Quantity."""

        return self.ureg.Quantity(data)

    @lru_cache()
    def conversion_factor(
        self, base_unit: Union[str, "quantity._Quantity"], conv_unit: Union[str, "quantity._Quantity"]
    ) -> float:
        """Provides the conversion factor from one unit to another.

        The conversion factor is based on the current contexts CODATA.

        Parameters
        ----------
        base_unit : Union[str, 'Quantity']
            The original units
        conv_unit : Union[str, 'Quantity']
            The units to convert to

        Examples
        --------

        >>> conversion_factor("meter", "picometer")
        1e-12

        >>> conversion_factor("feet", "meter")
        0.30479999999999996

        >>> conversion_factor(10 * ureg.feet, "meter")
        3.0479999999999996

        Returns
        -------
        float
            The requested conversion factor
        """

        # Add a little magic in case the incoming values have scalars

        from pint import quantity

        factor = 1.0

        # First make sure we either have Units or Quantities
        if isinstance(base_unit, str):
            base_unit = self.ureg.parse_expression(base_unit)

        if isinstance(conv_unit, str):
            conv_unit = self.ureg.parse_expression(conv_unit)

        # Need to play with prevector if we have Quantities
        if isinstance(base_unit, quantity._Quantity):
            factor *= base_unit.magnitude
            base_unit = base_unit.units

        if isinstance(conv_unit, quantity._Quantity):
            factor /= conv_unit.magnitude
            conv_unit = conv_unit.units

        return self.ureg.convert(factor, base_unit, conv_unit)

    def string_representation(self) -> str:
        """Print name, value, and units of all physical constants."""

        return print_variables(self.pc)


def run_comparison(context: str) -> None:
    """Compare the existing physical constant information for Psi4 (in checkup_data folder) to object instantiation from
    `context`. Specialized use."""

    self = PhysicalConstantsContext(context)

    try:
        from .. import checkup_data
    except ImportError:  # pragma: no cover
        print("Info for comparison (directory checkup_data) not installed. Run from source.")
        raise

    class bcolors:
        HEADER = "\033[95m"
        OKBLUE = "\033[94m"
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        BOLD = "\033[1m"
        UNDERLINE = "\033[4m"

    tol = 1.0e-8
    print(bcolors.OKBLUE + "\nChecking ({}) physconst vs. Psi4 ...".format(tol) + bcolors.ENDC)
    for pc in dir(checkup_data.physconst):
        if not pc.startswith("__"):
            ref = self.get(pc)
            val = getattr(checkup_data.physconst, pc)
            assert isinstance(ref, (int, float))
            rat = abs(1.0 - float(ref) / val)
            if rat > 1.0e-4:
                print(
                    bcolors.FAIL
                    + "Physical Constant {} ratio differs by {:12.8f}: {} (this) vs {} (psi)".format(pc, rat, ref, val)
                    + bcolors.ENDC
                )
            if rat > tol:
                print("Physical Constant {} ratio differs by {:12.8f}: {} (this) vs {} (psi)".format(pc, rat, ref, val))


def run_internal_comparison(old_context: str, new_context: str) -> None:
    """Compare two physical constant versions. Useful when adding new version."""

    self = PhysicalConstantsContext(old_context)
    other = PhysicalConstantsContext(new_context)

    class bcolors:
        HEADER = "\033[95m"
        OKBLUE = "\033[94m"
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        BOLD = "\033[1m"
        UNDERLINE = "\033[4m"

    both = set(self.pc.keys()).intersection(other.pc.keys())

    self_only = set(self.pc.keys()).difference(both)
    print(bcolors.WARNING + f"\nPhysConst in old {self.name} missing in new {other.name} ..." + bcolors.ENDC)
    for pc in sorted(self_only):
        print(pc, self.get(pc))

    other_only = set(other.pc.keys()).difference(both)
    print(bcolors.WARNING + f"\nPhysConst in new {other.name} missing in old {self.name} ..." + bcolors.ENDC)
    for pc in sorted(other_only):
        print(pc, other.get(pc))

    tol = 1.0e-8
    print(bcolors.OKBLUE + f"\nChecking ({tol}) physconst {self.name} vs. {other.name} ..." + bcolors.ENDC)
    for pc in sorted(both):
        if not pc.startswith("__"):
            ref = self.get(pc)
            val = other.get(pc)
            assert isinstance(ref, (int, float))
            rat = abs(1.0 - float(ref) / val)
            if rat > 1.0e-4:
                print(
                    bcolors.FAIL
                    + f"Physical Constant {pc} ratio differs by {rat:12.8f}: {ref} (this, {self.name}) vs {val} ({other.name})"
                    + bcolors.ENDC
                )
            if rat > tol:
                print(
                    f"Physical Constant {pc} ratio differs by {rat:12.8f}: {ref} (this, {self.name}) vs {val} ({other.name})"
                )

            refu = self.pc[pc].units
            valu = other.pc[pc].units
            if refu != valu:
                print(
                    bcolors.BOLD
                    + f"Physical Constant {pc} units differs: {refu} (this, {self.name}) vs {valu} ({other.name})"
                    + bcolors.ENDC
                )


def _get_pi(from_scratch: bool = False) -> "Decimal":
    """Get pi to 36 digits (or more with mpmath).

    Parameters
    ----------
    from_scratch : bool, optional
        If True, recomputes Pi from mpmath.

    Returns
    -------
    Decimal
        A representation of Pi
    """

    if from_scratch:  # pragma: no cover
        from mpmath import mp

        mp.dps = 36
        return mp.pi
    else:
        return Decimal("3.14159265358979323846264338327950288")


def write_c_header(context, filename="physconst.h"):
    """Write C header file defining physical constants and pi, all with ``pc_`` prefix."""

    self = PhysicalConstantsContext(context)

    pi = _get_pi(from_scratch=False)
    tau = 2 * pi

    text = [
        "#ifndef _qcelemental_physconst_h_",
        "#define _qcelemental_physconst_h_",
        "",
        f"/* This file is autogenerated from the QCElemental python module from {self.name} */",
        "",
        "/* clang-format off */",
        f"#define pc_pi {pi}",
        f"#define pc_twopi {tau}",
    ]

    for pc, qca in sorted(self.pc.items()):
        callname = qca.label.translate(self._transtable)
        noncomment = "#define pc_{} {}".format(callname, qca.data)
        text.append("{:80}  /*- {} [{}] {} -*/".format(noncomment, qca.label, qca.units, qca.comment))
    text.append("/* clang-format on */")

    text.append("")
    text.append("/* For Cray X1 compilers */")
    text.append("#ifndef M_PI")
    text.append("#define M_PI 3.14159265358979323846")
    text.append("#endif")
    text.append("")

    text.append("#endif /* header guard */")
    text.append("")

    with open(filename, "w") as handle:
        handle.write("\n".join(text))
    print("File written ({}). Remember to add license and clang-format it.".format(filename))


# singleton
constants = PhysicalConstantsContext("CODATA2014")



from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Set, Union

import numpy as np
from pydantic import Field, constr, validator

from ..util import provenance_stamp
from .basemodels import ProtoModel
from .basis import BasisSet
from .common_models import ComputeError, DriverEnum, Model, Provenance, qcschema_input_default, qcschema_output_default
from .molecule import Molecule
from .types import Array

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs


class AtomicResultProperties(ProtoModel):
    """
    Named properties of quantum chemistry computations following the MolSSI QCSchema.
    """

    # Calcinfo
    calcinfo_nbasis: Optional[int] = Field(None, description="The number of basis functions for the computation.")
    calcinfo_nmo: Optional[int] = Field(None, description="The number of molecular orbitals for the computation.")
    calcinfo_nalpha: Optional[int] = Field(None, description="The number of alpha electrons in the computation.")
    calcinfo_nbeta: Optional[int] = Field(None, description="The number of beta electrons in the computation.")
    calcinfo_natom: Optional[int] = Field(None, description="The number of atoms in the computation.")

    # Canonical
    nuclear_repulsion_energy: Optional[float] = Field(None, description="The nuclear repulsion energy energy.")
    return_energy: Optional[float] = Field(
        None, description="The energy of the requested method, identical to `return_value` for energy computations."
    )

    # SCF Keywords
    scf_one_electron_energy: Optional[float] = Field(
        None, description="The one-electron (core Hamiltonian) energy contribution to the total SCF energy."
    )
    scf_two_electron_energy: Optional[float] = Field(
        None, description="The two-electron energy contribution to the total SCF energy."
    )
    scf_vv10_energy: Optional[float] = Field(
        None, description="The VV10 functional energy contribution to the total SCF energy."
    )
    scf_xc_energy: Optional[float] = Field(
        None, description="The functional (XC) energy contribution to the total SCF energy."
    )
    scf_dispersion_correction_energy: Optional[float] = Field(
        None,
        description="The dispersion correction appended to an underlying functional when a DFT-D method is requested.",
    )
    scf_dipole_moment: Optional[Array[float]] = Field(None, description="The X, Y, and Z dipole components.")
    scf_quadrupole_moment: Optional[Array[float]] = Field(
        None, description="The (3, 3) quadrupole components (redundant; 6 unique)."
    )
    scf_total_energy: Optional[float] = Field(
        None, description="The total electronic energy of the SCF stage of the calculation."
    )
    scf_iterations: Optional[int] = Field(None, description="The number of SCF iterations taken before convergence.")

    # MP2 Keywords
    mp2_same_spin_correlation_energy: Optional[float] = Field(
        None, description="The portion of MP2 doubles correlation energy from same-spin (i.e. triplet) correlations."
    )
    mp2_opposite_spin_correlation_energy: Optional[float] = Field(
        None,
        description="The portion of MP2 doubles correlation energy from opposite-spin (i.e. singlet) correlations.",
    )
    mp2_singles_energy: Optional[float] = Field(
        None, description="The singles portion of the MP2 correlation energy. Zero except in ROHF."
    )
    mp2_doubles_energy: Optional[float] = Field(
        None,
        description="The doubles portion of the MP2 correlation energy including same-spin and opposite-spin correlations.",
    )
    mp2_total_correlation_energy: Optional[float] = Field(
        None, description="The MP2 correlation energy."
    )  # Old name, to be deprecated
    mp2_correlation_energy: Optional[float] = Field(None, description="The MP2 correlation energy.")
    mp2_total_energy: Optional[float] = Field(
        None, description="The total MP2 energy (MP2 correlation energy + HF energy)."
    )
    mp2_dipole_moment: Optional[Array[float]] = Field(None, description="The MP2 X, Y, and Z dipole components.")

    # CCSD Keywords
    ccsd_same_spin_correlation_energy: Optional[float] = Field(
        None, description="The portion of CCSD doubles correlation energy from same-spin (i.e. triplet) correlations."
    )
    ccsd_opposite_spin_correlation_energy: Optional[float] = Field(
        None,
        description="The portion of CCSD doubles correlation energy from opposite-spin (i.e. singlet) correlations",
    )
    ccsd_singles_energy: Optional[float] = Field(
        None, description="The singles portion of the CCSD correlation energy. Zero except in ROHF."
    )
    ccsd_doubles_energy: Optional[float] = Field(
        None,
        description="The doubles portion of the CCSD correlation energy including same-spin and opposite-spin correlations.",
    )
    ccsd_correlation_energy: Optional[float] = Field(None, description="The CCSD correlation energy.")
    ccsd_total_energy: Optional[float] = Field(
        None, description="The total CCSD energy (CCSD correlation energy + HF energy)."
    )
    ccsd_dipole_moment: Optional[Array[float]] = Field(None, description="The CCSD X, Y, and Z dipole components.")
    ccsd_iterations: Optional[int] = Field(None, description="The number of CCSD iterations taken before convergence.")

    # CCSD(T) keywords
    ccsd_prt_pr_correlation_energy: Optional[float] = Field(None, description="The CCSD(T) correlation energy.")
    ccsd_prt_pr_total_energy: Optional[float] = Field(
        None, description="The total CCSD(T) energy (CCSD(T) correlation energy + HF energy)."
    )
    ccsd_prt_pr_dipole_moment: Optional[Array[float]] = Field(
        None, description="The CCSD(T) X, Y, and Z dipole components."
    )

    class Config(ProtoModel.Config):
        force_skip_defaults = True

    def __repr_args__(self) -> "ReprArgs":
        return [(k, v) for k, v in self.dict().items()]

    @validator(
        "scf_dipole_moment",
        "mp2_dipole_moment",
        "ccsd_dipole_moment",
        "ccsd_prt_pr_dipole_moment",
        "scf_quadrupole_moment",
    )
    def _validate_poles(cls, v, values, field):
        if v is None:
            return v

        if field.name.endswith("_dipole_moment"):
            order = 1
        elif field.name.endswith("_quadrupole_moment"):
            order = 2

        shape = tuple([3] * order)
        return np.asarray(v).reshape(shape)

    def dict(self, *args, **kwargs):
        # pure-json dict repr for QCFractal compliance, see https://github.com/MolSSI/QCFractal/issues/579
        kwargs["encoding"] = "json"
        return super().dict(*args, **kwargs)


class WavefunctionProperties(ProtoModel):

    # Class properties
    _return_results_names: Set[str] = {
        "orbitals_a",
        "orbitals_b",
        "density_a",
        "density_b",
        "fock_a",
        "fock_b",
        "eigenvalues_a",
        "eigenvalues_b",
        "occupations_a",
        "occupations_b",
    }

    # The full basis set description of the quantities
    basis: BasisSet = Field(..., description=str(BasisSet.__doc__))
    restricted: bool = Field(
        ...,
        description=str(
            "If the computation was restricted or not (alpha == beta). If True, all beta quantities are skipped."
        ),
    )

    # Core Hamiltonian
    h_core_a: Optional[Array[float]] = Field(None, description="Alpha-spin core (one-electron) Hamiltonian.")
    h_core_b: Optional[Array[float]] = Field(None, description="Beta-spin core (one-electron) Hamiltonian.")
    h_effective_a: Optional[Array[float]] = Field(
        None, description="Alpha-spin effective core (one-electron) Hamiltonian."
    )
    h_effective_b: Optional[Array[float]] = Field(
        None, description="Beta-spin effective core (one-electron) Hamiltonian "
    )

    # SCF Results
    scf_orbitals_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin orbitals.")
    scf_orbitals_b: Optional[Array[float]] = Field(None, description="SCF beta-spin orbitals.")
    scf_density_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin density matrix.")
    scf_density_b: Optional[Array[float]] = Field(None, description="SCF beta-spin density matrix.")
    scf_fock_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin Fock matrix.")
    scf_fock_b: Optional[Array[float]] = Field(None, description="SCF beta-spin Fock matrix.")
    scf_eigenvalues_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin eigenvalues.")
    scf_eigenvalues_b: Optional[Array[float]] = Field(None, description="SCF beta-spin eigenvalues.")
    scf_occupations_a: Optional[Array[float]] = Field(None, description="SCF alpha-spin occupations.")
    scf_occupations_b: Optional[Array[float]] = Field(None, description="SCF beta-spin occupations.")

    # Return results, must be defined last
    orbitals_a: Optional[str] = Field(None, description="Index to the alpha-spin orbitals of the primary return.")
    orbitals_b: Optional[str] = Field(None, description="Index to the beta-spin orbitals of the primary return.")
    density_a: Optional[str] = Field(None, description="Index to the alpha-spin density of the primary return.")
    density_b: Optional[str] = Field(None, description="Index to the beta-spin density of the primary return.")
    fock_a: Optional[str] = Field(None, description="Index to the alpha-spin Fock matrix of the primary return.")
    fock_b: Optional[str] = Field(None, description="Index to the beta-spin Fock matrix of the primary return.")
    eigenvalues_a: Optional[str] = Field(None, description="Index to the alpha-spin eigenvalues of the primary return.")
    eigenvalues_b: Optional[str] = Field(None, description="Index to the beta-spin eigenvalues of the primary return.")
    occupations_a: Optional[str] = Field(
        None, description="Index to the alpha-spin orbital eigenvalues of the primary return."
    )
    occupations_b: Optional[str] = Field(
        None, description="Index to the beta-spin orbital eigenvalues of the primary return."
    )

    class Config(ProtoModel.Config):
        force_skip_defaults = True

    @validator("scf_eigenvalues_a", "scf_eigenvalues_b", "scf_occupations_a", "scf_occupations_b")
    def _assert1d(cls, v, values):

        try:
            v = v.reshape(-1)
        except (ValueError, AttributeError):
            raise ValueError("Vector must be castable to shape (-1, )!")
        return v

    @validator("scf_orbitals_a", "scf_orbitals_b")
    def _assert2d_nao_x(cls, v, values):
        bas = values.get("basis", None)

        # Do not raise multiple errors
        if bas is None:
            return v

        try:
            v = v.reshape(bas.nbf, -1)
        except (ValueError, AttributeError):
            raise ValueError("Matrix must be castable to shape (nbf, -1)!")
        return v

    @validator(
        "h_core_a",
        "h_core_b",
        "h_effective_a",
        "h_effective_b",
        # SCF
        "scf_density_a",
        "scf_density_b",
        "scf_fock_a",
        "scf_fock_b",
    )
    def _assert2d(cls, v, values):
        bas = values.get("basis", None)

        # Do not raise multiple errors
        if bas is None:
            return v

        try:
            v = v.reshape(bas.nbf, bas.nbf)
        except (ValueError, AttributeError):
            raise ValueError("Matrix must be castable to shape (nbf, nbf)!")
        return v

    @validator(
        "orbitals_a",
        "orbitals_b",
        "density_a",
        "density_b",
        "fock_a",
        "fock_b",
        "eigenvalues_a",
        "eigenvalues_b",
        "occupations_a",
        "occupations_b",
    )
    def _assert_exists(cls, v, values):

        if values.get(v, None) is None:
            raise ValueError(f"Return quantity {v} does not exist in the values.")
        return v


class WavefunctionProtocolEnum(str, Enum):
    """
    Wavefunction to keep from a Result computation.
    """

    all = "all"
    orbitals_and_eigenvalues = "orbitals_and_eigenvalues"
    return_results = "return_results"
    none = "none"


class ErrorCorrectionProtocol(ProtoModel):
    """Configuration for how QCEngine handles error correction

    WARNING: These protocols are currently experimental and only supported by NWChem tasks
    """

    default_policy: bool = Field(
        True, description="Whether to allow error corrections to be used " "if not directly specified in `policies`"
    )
    # TODO (wardlt): Consider support for common policies (e.g., 'only increase iterations') as strings (see #182)
    policies: Optional[Dict[str, bool]] = Field(
        None,
        description="Settings that define whether specific error corrections are allowed. "
        "Keys are the name of a known error and values are whether it is allowed to be used.",
    )

    def allows(self, policy: str):
        if self.policies is None:
            return self.default_policy
        return self.policies.get(policy, self.default_policy)


class AtomicResultProtocols(ProtoModel):
    """
    Protocols regarding the manipulation of a Result output data.
    """

    wavefunction: WavefunctionProtocolEnum = Field(
        WavefunctionProtocolEnum.none, description=str(WavefunctionProtocolEnum.__doc__)
    )
    stdout: bool = Field(True, description="Primary output file to keep from a Result computation")
    error_correction: ErrorCorrectionProtocol = Field(
        ErrorCorrectionProtocol(), description="Policies for error correction"
    )

    class Config:
        force_skip_defaults = True


### Primary models


class AtomicInput(ProtoModel):
    """The MolSSI Quantum Chemistry Schema"""

    id: Optional[str] = Field(None, description="An optional ID of the ResultInput object.")
    schema_name: constr(strip_whitespace=True, regex=qcschema_input_default) = qcschema_input_default  # type: ignore
    schema_version: int = 1

    molecule: Molecule = Field(..., description="The molecule to use in the computation.")
    driver: DriverEnum = Field(..., description=str(DriverEnum.__doc__))
    model: Model = Field(..., description=str(Model.__base_doc__))
    keywords: Dict[str, Any] = Field({}, description="The program specific keywords to be used.")
    protocols: AtomicResultProtocols = Field(
        AtomicResultProtocols(), description=str(AtomicResultProtocols.__base_doc__)
    )

    extras: Dict[str, Any] = Field({}, description="Extra fields that are not part of the schema.")

    provenance: Provenance = Field(Provenance(**provenance_stamp(__name__)), description=str(Provenance.__base_doc__))

    def __repr_args__(self) -> "ReprArgs":
        return [
            ("driver", self.driver.value),
            ("model", self.model.dict()),
            ("molecule_hash", self.molecule.get_hash()[:7]),
        ]


class AtomicResult(AtomicInput):
    schema_name: constr(strip_whitespace=True, regex=qcschema_output_default) = qcschema_output_default  # type: ignore

    properties: AtomicResultProperties = Field(..., description=str(AtomicResultProperties.__base_doc__))
    wavefunction: Optional[WavefunctionProperties] = Field(None, description=str(WavefunctionProperties.__base_doc__))

    return_result: Union[float, Array[float], Dict[str, Any]] = Field(
        ..., description="The value requested by the 'driver' attribute."
    )  # type: ignore

    stdout: Optional[str] = Field(None, description="The standard output of the program.")
    stderr: Optional[str] = Field(None, description="The standard error of the program.")

    success: bool = Field(
        ..., description="The success of a given programs execution. If False, other fields may be blank."
    )
    error: Optional[ComputeError] = Field(None, description=str(ComputeError.__base_doc__))
    provenance: Provenance = Field(..., description=str(Provenance.__base_doc__))

    @validator("schema_name", pre=True)
    def _input_to_output(cls, v):
        """If qcschema_input is passed in, cast it to output, otherwise no"""
        if v.lower().strip() in [qcschema_input_default, qcschema_output_default]:
            return qcschema_output_default
        raise ValueError(
            "Only {0} or {1} is allowed for schema_name, "
            "which will be converted to {0}".format(qcschema_output_default, qcschema_input_default)
        )

    @validator("return_result")
    def _validate_return_result(cls, v, values):
        if values["driver"] == "gradient":
            v = np.asarray(v).reshape(-1, 3)
        elif values["driver"] == "hessian":
            v = np.asarray(v)
            nsq = int(v.size ** 0.5)
            v.shape = (nsq, nsq)

        return v

    @validator("wavefunction", pre=True)
    def _wavefunction_protocol(cls, value, values):

        # We are pre, gotta do extra checks
        if value is None:
            return value
        elif isinstance(value, dict):
            wfn = value.copy()
        elif isinstance(value, WavefunctionProperties):
            wfn = value.dict()
        else:
            raise ValueError("wavefunction must be None, a dict, or a WavefunctionProperties object.")

        # Do not propagate validation errors
        if "protocols" not in values:
            raise ValueError("Protocols was not properly formed.")

        # Handle restricted
        restricted = wfn.get("restricted", None)
        if restricted is None:
            raise ValueError("`restricted` is required.")

        if restricted:
            for k in list(wfn.keys()):
                if k.endswith("_b"):
                    wfn.pop(k)

        # Handle protocols
        wfnp = values["protocols"].wavefunction
        return_keep = None
        if wfnp == "all":
            pass
        elif wfnp == "none":
            wfn = None
        elif wfnp == "return_results":
            return_keep = [
                "orbitals_a",
                "orbitals_b",
                "density_a",
                "density_b",
                "fock_a",
                "fock_b",
                "eigenvalues_a",
                "eigenvalues_b",
                "occupations_a",
                "occupations_b",
            ]
        elif wfnp == "orbitals_and_eigenvalues":
            return_keep = ["orbitals_a", "orbitals_b", "eigenvalues_a", "eigenvalues_b"]
        else:
            raise ValueError(f"Protocol `wavefunction:{wfnp}` is not understood.")

        if return_keep is not None:
            ret_wfn = {"restricted": restricted}
            if "basis" in wfn:
                ret_wfn["basis"] = wfn["basis"]

            for rk in return_keep:
                key = wfn.get(rk, None)
                if key is None:
                    continue

                ret_wfn[rk] = key
                ret_wfn[key] = wfn[key]

            return ret_wfn
        else:
            return wfn

    @validator("stdout")
    def _stdout_protocol(cls, value, values):

        # Do not propagate validation errors
        if "protocols" not in values:
            raise ValueError("Protocols was not properly formed.")

        outp = values["protocols"].stdout
        if outp is True:
            return value
        elif outp is False:
            return None
        else:
            raise ValueError(f"Protocol `stdout:{outp}` is not understood")


class ResultProperties(AtomicResultProperties):
    def __init__(self, *args, **kwargs):
        from warnings import warn

        warn(
            "ResultProperties has been renamed to AtomicResultProperties and will be removed in v0.13.0",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class ResultProtocols(AtomicResultProtocols):
    def __init__(self, *args, **kwargs):
        from warnings import warn

        warn(
            "ResultProtocols has been renamed to AtomicResultProtocols and will be removed in v0.13.0",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class ResultInput(AtomicInput):
    def __init__(self, *args, **kwargs):
        from warnings import warn

        warn("ResultInput has been renamed to AtomicInput and will be removed in v0.13.0", DeprecationWarning)
        super().__init__(*args, **kwargs)


class Result(AtomicResult):
    def __init__(self, *args, **kwargs):
        from warnings import warn

        warn("Result has been renamed to AtomicResult and will be removed in v0.13.0", DeprecationWarning)
        super().__init__(*args, **kwargs)



try:
    import pydantic
except ImportError:  # pragma: no cover
    raise ImportError(
        "Python module pydantic not found. Solve by installing it: "
        "`conda install pydantic -c conda-forge` or `pip install pydantic`"
    )

from . import types
from .align import AlignmentMill
from .basemodels import AutodocBaseSettings, ProtoModel
from .basis import BasisSet
from .common_models import ComputeError, DriverEnum, FailedOperation, Provenance
from .molecule import Molecule
from .procedures import Optimization, OptimizationInput, OptimizationResult
from .results import AtomicInput, AtomicResult, AtomicResultProperties, Result, ResultInput, ResultProperties



from .alvarez_2008_covalent_radii import alvarez_2008_covalent_radii
from .mantina_2009_vanderwaals_radii import mantina_2009_vanderwaals_radii
from .nist_2011_atomic_weights import nist_2011_atomic_weights
from .nist_2014_codata import nist_2014_codata
from .nist_2018_codata import nist_2018_codata



import os
from decimal import Decimal

import pytest

import qcelemental


@pytest.mark.parametrize("inp", ["He100", "-1", -1, -1.0, "cat", 200, "Cr_highspin"])
def test_id_resolution_error_bad_element(inp):
    with pytest.raises(qcelemental.NotAnElementError):
        qcelemental.covalentradii.get(inp)


@pytest.mark.parametrize("inp", ["X", "Bk", 100])
def test_id_resolution_error(inp):
    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.covalentradii.get(inp)

    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.covalentradii.get(inp, return_tuple=True)

    with pytest.raises(qcelemental.DataUnavailableError):
        qcelemental.covalentradii.get(inp, return_tuple=True, missing=3.0)

    assert qcelemental.covalentradii.get(inp, missing=4.0) == pytest.approx(4.0, 1.0e-9)


a2b = 1.0 / qcelemental.constants.bohr2angstroms


@pytest.mark.parametrize(
    "inp,expected",
    [
        # Kr 84
        ("KRYPTON", 1.16),
        ("kr", 1.16),
        ("kr84", 1.16),
        (36, 1.16),
        # aliases
        ("C", 0.76),
        ("C_sp", 0.69),
        ("MN", 1.61),
        ("Mn_lowspin", 1.39),
        # Deuterium
        ("D", 0.31),
        ("h2", 0.31),
    ],
)
def test_get(inp, expected):
    assert qcelemental.covalentradii.get(inp, units="angstrom") == pytest.approx(expected, 1.0e-9)
    assert qcelemental.covalentradii.get(inp) == pytest.approx(a2b * expected, 1.0e-9)


def test_get_tuple():
    ref = {"label": "Mn", "units": "angstrom", "data": Decimal("1.61")}
    dqca = qcelemental.covalentradii.get("manganese", return_tuple=True).dict()

    for itm in ref:
        assert ref[itm] == dqca[itm]


def test_c_header():
    qcelemental.covalentradii.write_c_header("header.h")
    os.remove("header.h")


def test_representation():
    qcelemental.covalentradii.string_representation()


def test_str():
    assert "CovalentRadii(" in str(qcelemental.covalentradii)


def test_covradmaker2018():
    with pytest.raises(KeyError) as e:
        qcelemental.CovalentRadii("COVRADMAKER2018")

    assert "only contexts {'ALVAREZ2008', } are currently supported" in str(e.value)



"""
Contains metadata about Processors
"""

import difflib
import re
from enum import Enum
from functools import lru_cache
from typing import List, Optional

from pydantic import Field

from ..models import ProtoModel


class VendorEnum(str, Enum):
    """Allowed processor vendors, used for validation."""

    amd = "amd"
    intel = "intel"
    nvidia = "nvidia"
    arm = "arm"


class InstructionSetEnum(int, Enum):
    """Allowed instruction sets for CPUs in an ordinal enum."""

    none = 0
    sse = 1
    avx = 2
    avx2 = 3
    avx512 = 4


class ProcessorInfo(ProtoModel):
    name: str = Field(..., description="Name of the processor.")
    ncores: int = Field(..., description="The number of physical cores on the chip.")
    nthreads: Optional[int] = Field(..., description="The maximum number of concurrent threads.")
    base_clock: float = Field(..., description="The base clock frequency (GHz).")
    boost_clock: Optional[float] = Field(..., description="The boost clock frequency (GHz).")
    model: str = Field(..., description="The model number of the chip.")
    family: str = Field(..., description="The family of the chip.")
    launch_date: Optional[int] = Field(..., description="The launch year of the chip.")
    target_use: str = Field(..., description="Target use case (Desktop, Server, etc).")
    vendor: VendorEnum = Field(..., description="The vendor the chip is produced by.")
    microarchitecture: Optional[str] = Field(..., description="The microarchitecture the chip follows.")
    instructions: InstructionSetEnum = Field(..., description="The maximum vectorized instruction set available.")
    type: str = Field(..., description="The type of chip (cpu, gpu, etc).")


class ProcessorContext:
    """Information about Process information.

    Parameters
    ----------
    context : {'default'}
        Origin of loaded data.

    Attributes
    ----------
    processors : List[ProcessorInfo]
        A list of all processors known
    index : Dict[(str, str): ProcessorInfo]
        A (vendor, model) index to the processor information
    index_vendor : Dict[str, Dict[str, ProcessorInfo]]
        A vendor, model nested dictionary index for processor information
    name : str
        The name of the context ('default')

    """

    def __init__(self, context: str = "defualt"):
        if context != "default":
            raise KeyError(f"Context set as '{context}'," + " only contexts {'default'} are currently supported")

        from .data import cpu_data_blob

        self.processors: List[ProcessorInfo] = [ProcessorInfo(**data) for data in cpu_data_blob.data_blob]

        self.index = {(x.vendor, self.process_names(x.model)): x for x in self.processors}

        self.index_vendor = {k.name: {} for k in VendorEnum}
        for proc in self.processors:
            self.index_vendor[proc.vendor][self.process_names(proc.model)] = proc

        self.name = context

    def __str__(self) -> str:
        return "ProcessorContext(context='{}')".format(self.name)

    def process_names(self, name):
        name = name.lower()
        name = name.replace("(tm)", "").replace("(r)", "").replace("™", "")
        name = " ".join(name.split())
        name = name.strip()
        return name


context = ProcessorContext("default")


@lru_cache(maxsize=1024)
def get(name: str, vendor=None, cutoff=0.9) -> ProcessorInfo:

    name = context.process_names(name.split("@")[0])

    if ("amd" in name) or (vendor == "amd"):
        vendor = "amd"

        name = re.split(r"\d+-core pro", name)[0]

        # Opteron model names are only numbers?
        if "opteron" in name:
            name = name.replace("processor", "")
            name = name.replace("opteron", "")
            name = name.replace("amd", "")

    elif "intel" in name or (vendor == "intel"):
        vendor = "intel"

        if "xeon" in name:
            name = name.replace(" v", "v")
            if name.endswith(" 0"):
                name = name[:-2]

        name = name.replace("intel", "").replace("cpu", "")
        name = name.replace("core", "").replace("xeon", "")
        name = name.replace("silver", "").replace("gold", "").replace("phi", "")

    if vendor is None:
        raise KeyError("Could not determine vendor, please provide one.")

    index = context.index_vendor[vendor]

    name = context.process_names(name)
    matches = difflib.get_close_matches(name, index.keys(), cutoff=cutoff)

    if matches:
        return index[matches[0]]
    else:
        raise KeyError(f"Could not find processor {vendor}: {name}.")


def list_names():
    return sorted([cpu.name for cpu in context.processors])



import pytest

from qcelemental.info import cpu_info, dft_info


@pytest.mark.parametrize(
    "functional,base_name",
    [
        ("svwn", "svwn"),
        ("b3lyp", "b3lyp"),
        ("b3lyp-d3", "b3lyp"),
        ("b3lyp-d3zero", "b3lyp"),
        ("b3lyp-d3mbj", "b3lyp"),
        ("b3lyp-nl", "b3lyp"),
        ("dsd-blyp", "dsd-blyp"),
        ("core-dsd-blyp", "core-dsd-blyp"),
        ("core-dsd-blyp-d3bj", "core-dsd-blyp"),
        ("pbe0-dh", "pbe0-dh"),
        ("m06-2x-d3bj", "m06-2x"),
    ],
)
def test_dft_info_get(functional, base_name):
    assert dft_info.get(functional).name == base_name


def test_dft_info_data():
    dft = dft_info.get("core-dsd-blyp")
    assert dft.name == "core-dsd-blyp"
    assert dft.deriv == 1
    assert dft.c_hybrid is True
    assert dft.x_hybrid is True


def test_cpu_info_index_lengths():
    assert len(cpu_info.context.processors) == len(cpu_info.context.index), "Duplicates found in Index."


@pytest.mark.parametrize(
    "name, model",
    [
        ("AMD Ryzen Threadripper 1950X 16-Core Processor", "AMD Ryzen™ Threadripper™ 1950X"),
        ("AMD Opteron(tm) Processor 6376", "6376"),
        ("AMD EPYC 7601 32-Core Processor", "AMD EPYC™ 7601"),
        ("AMD EPYC 7551P 32-Core Processor", "AMD EPYC™ 7551P"),
        ("Intel(R) Xeon Phi(TM) CPU 7230 @ 1.30GHz", "7230"),
        ("Intel(R) Xeon(R) Gold 6130 CPU @ 2.10GHz", "6130"),
        ("Intel(R) Xeon(R) Silver 4214 CPU @ 2.20GHz", "4214"),
        ("Intel(R) Xeon(R) CPU E5-2683 v4 @ 2.10GHz", "E5-2683V4"),
        ("Intel(R) Xeon(R) CPU           X5660  @ 2.80GHz", "X5660"),
        ("Intel(R) Xeon(R) CPU E7-8867 v4 @ 2.40GHz", "E7-8867V4"),
        ("Intel(R) Xeon(R) CPU E5-2670 0 @ 2.60GHz", "E5-2670"),
        ("Intel(R) Xeon(R) CPU E5-2660 0 @ 2.60GHz", "E5-2660"),
        ("Intel(R) Xeon(R) CPU E5-2690 0 @ 2.60GHz", "E5-2690"),
        ("AMD Opteron(tm) Processor 6168", "6168"),
        ("AMD Opteron(tm) Processor 6174", "6174"),
        ("Intel(R) Xeon(R) CPU E5-2673 v3 @ 2.40GHz", "E5-2673V3"),
        ("Intel(R) Core(TM) i7-7820HQ CPU @ 2.90GHz", "i7-7820HQ"),
        ("Intel(R) Core(TM) i5-7360U CPU @ 2.30GHz", "i5-7360U"),
        ("Intel(R) Core(TM) i5-8279U CPU @ 2.40GHz", "i5-8279U"),
    ],
)
def test_cpu_info_search(name, model):

    cpu = cpu_info.get(name)
    assert cpu is not None, name
    assert cpu.model == model, name


def test_cpu_info_errors():

    with pytest.raises(KeyError) as exc:
        cpu_info.get("E7-8867 V4")

    assert "not determine vendor" in exc.value.args[0], exc.value.args[0]

    with pytest.raises(KeyError) as exc:
        cpu_info.get("Intel E0-9999 V4")

    assert "intel: e0-9999" in exc.value.args[0], exc.value.args[0]


def test_cpu_info_matches():
    assert cpu_info.get("E7-8867 V4", vendor="intel").model == "E7-8867V4"


@pytest.mark.parametrize(
    "name, fixes",
    [
        ("AMD EPYC 7551P 32-Core Processor", {"launch_date": 2017}),
        ("AMD Opteron(tm) Processor 6376", {"launch_date": 2012}),
        ("AMD Opteron(TM) Processor 6276", {"launch_date": 2011}),
        ("AMD EPYC 7401P 24-Core Processor", {"launch_date": 2017}),
        ("AMD EPYC 7601 32-Core Processor", {"launch_date": 2017}),
        ("AMD Opteron(TM) Processor 6274", {"launch_date": 2011}),
    ],
)
def test_cpu_fixes(name, fixes):
    info = cpu_info.get(name)
    print(info)
    for k, v in fixes.items():
        assert getattr(info, k) == v



"""
Contains van der Waals radii
"""

import collections
from decimal import Decimal
from typing import Dict, Union

from .datum import Datum, print_variables
from .exceptions import DataUnavailableError
from .periodic_table import periodictable


class VanderWaalsRadii:
    """Van der Waals radii sets.

    Parameters
    ----------
    context : {'MANTINA2009'}
        Origin of loaded data.

    Attributes
    ----------
    vdwr : dict of Datum
        Each van der Waals radius is an entry in `vdwr`, where key is the
        "Fe"-cased element symbol if generic or symbol-prefixed label
        if specialized within element. The value is a Datum object with
        `lbl` the same as key, `units` and `data` value as Decimal object.
    doi : str
        The DOI of the current context.
    name : str
        The name of the context ('MANTINA2009')
    native_units : str
        The units the original data was provided in.
    year : int
        The year the context was created.

    """

    def __init__(self, context: str = "MANTINA2009"):
        self.vdwr: Dict[str, Datum] = collections.OrderedDict()

        from .data import mantina_2009_vanderwaals_radii

        if context == "MANTINA2009":
            self.doi = mantina_2009_vanderwaals_radii["doi"]
            self.native_units = mantina_2009_vanderwaals_radii["units"]

            # TypedDict wont be in until 3.8, have to ignore heterogeneous dicts for now
            for vdwr in mantina_2009_vanderwaals_radii["vanderwaals_radii"]:  # type: ignore
                self.vdwr[vdwr[0]] = Datum(vdwr[0], self.native_units, Decimal(vdwr[1]), doi=self.doi)
        else:
            raise KeyError("Context set as '{}', only contexts {'MANTINA2009', } are currently supported")

        self.name = context
        self.year = int(mantina_2009_vanderwaals_radii["date"][:4])  # type: ignore

    def __str__(self) -> str:
        return "VanderWaalsRadii(context='{}')".format(self.name)

    def get(
        self, atom: Union[int, str], *, return_tuple: bool = False, units: str = "bohr", missing: float = None
    ) -> Union[float, "Datum"]:  # lgtm [py/similar-function]
        """
        Access a van der Waals radius for species ``atom``.

        Parameters
        ----------
        atom : int or str
            Identifier for element or nuclide, e.g., ``H``, ``C``, ``Al``.
        units : str, optional
            Units of returned value. To return in native unit (MANTINA2009: angstrom), pass it explicitly.
            Only relevant for ``return_tuple=False`` since ``True`` returns underlying data structure with native units.
        missing : float or None, optional
            How to handle when ``atom`` is valid but outside the available data range. When ``None``, raises DataUnavailableError.
            When a float, returns that float, so supply in ``units`` units. Supplying a float is a more compact assurance
            that a call will work over all the periodic table than the equivalent

            .. code-block:: python

                try:
                    rad = qcel.vdwradii.get(atom)
                except qcel.DataUnavailableError:
                    rad = 4.0

            Only relevant for ``return_tuple=False``.
        return_tuple : bool, optional
            See below.

        Returns
        -------
        float
            When ``return_tuple=False``, value of Van der Waals radius. If multiple defined for element, returns largest.
        qcelemental.Datum
            When ``return_tuple=True``, Datum with units, description, uncertainty, and value of van der Waals radius as Decimal (preserving significant figures).
            If multiple defined for element, returns largest.

        Raises
        ------
        NotAnElementError
            If `atom` cannot be resolved into an element or nuclide or label.
        DataUnavailableError
            If `atom` is a valid element or nuclide but not one for which a van der Waals radius is available and `missing=None`.

        """
        if atom in self.vdwr.keys():
            # catch extra labels like 'C_sp3'
            identifier = atom
        else:
            identifier = periodictable.to_E(atom)

        try:
            assert isinstance(identifier, str)  # Should be string by now
            qca = self.vdwr[identifier]
        except KeyError as e:
            if missing is not None and return_tuple is False:
                return missing
            else:
                raise DataUnavailableError("vanderwaals radius", identifier) from e

        if return_tuple:
            return qca
        else:
            return qca.to_units(units)

    def string_representation(self) -> str:
        """Print name, value, and units of all van der Waals radii."""

        return print_variables(self.vdwr)

    def write_c_header(self, filename: str = "vdwrad.h", missing: float = 2.0) -> None:  # lgtm [py/similar-function]
        """Write C header file defining Van der Waals radii array.

        Parameters
        ----------
        filename : str, optional
            File name for header. Note that changing this won't change the header guard.
        missing : float, optional
            In order that the C array be atomic-number indexable and that it span the
            periodic table, this value is used anywhere data is missing.

        """
        text = [
            "#ifndef _qcelemental_vdwrad_h_",
            "#define _qcelemental_vdwrad_h_",
            "",
            "/* This file is autogenerated from the QCElemental python module */",
            "",
            "const double vanderwaals_radii[] = {",
        ]

        for el in periodictable.E:
            try:
                qca = self.vdwr[el]
                text.append("{},  /*- [{}] {} {} -*/".format(qca.data, qca.units, qca.label, qca.comment))
            except KeyError:
                text.append(
                    "{:.2f},  /*- [{}] {} {} -*/".format(
                        missing, self.native_units, el, "Default value for missing data"
                    )
                )

        text.append("};")
        text.append("#endif /* header guard */")
        text.append("")

        with open(filename, "w") as handle:
            handle.write("\n".join(text))
        print("File written ({}). Remember to add license and clang-format it.".format(filename))


# singleton
vdwradii = VanderWaalsRadii("MANTINA2009")



import os
from decimal import Decimal

import pytest

import qcelemental


@pytest.fixture(scope="module", params=[None, "CODATA2014", "CODATA2018"])
def contexts(request):
    if request.param:
        constants = qcelemental.PhysicalConstantsContext(request.param)
        context = request.param
    else:
        constants = qcelemental.constants
        context = constants.name

    if context == "CODATA2014":
        return (constants, "27.21138602", "10.18434/T4WW24", "uncertainty=0.000 000 17")
    elif context == "CODATA2018":
        return (constants, "27.211386245988", "", "uncertainty=0.000 000 000 053")


def test_access_1a(contexts):
    constants, _, _, _ = contexts
    assert constants.c == 299792458


def test_access_1b(contexts):
    constants, _, _, _ = contexts
    assert constants.speed_of_light_in_vacuum == 299792458


def test_access_1c(contexts):
    constants, _, _, _ = contexts
    assert constants.pc["speed of light in vacuum"].data == 299792458


def test_access_1d(contexts):
    constants, _, _, _ = contexts
    assert constants.get("speed of light in vacuum") == 299792458


def test_access_1e(contexts, request):
    constants, _, doi, _ = contexts

    qca = constants.get("speed of light in vacuum", return_tuple=True)

    assert qca.units == "m s^{-1}"
    assert qca.comment == "uncertainty=(exact)"
    assert qca.doi == doi
    assert qca.data == 299792458


def test_access_1f(contexts):
    constants, _, _, _ = contexts
    assert constants.get("speed of LIGHT in vacuum") == 299792458


def test_access_2a(contexts):
    constants, ans, _, _ = contexts
    assert constants.Hartree_energy_in_eV == float(ans)


def test_access_2b(contexts):
    constants, ans, _, _ = contexts
    assert constants.hartree2ev == float(ans)


def test_access_2c(contexts):
    constants, ans, _, _ = contexts
    assert constants.pc["hartree energy in ev"].data == Decimal(ans)


def test_access_2d(contexts):
    constants, ans, _, _ = contexts
    assert constants.get("hartree energy in ev") == float(ans)


def test_access_2e(contexts):
    constants, ans, doi, comment = contexts

    qca = constants.get("Hartree energy in eV", return_tuple=True)
    print(qca)

    assert qca.units == "eV"
    assert qca.comment == comment
    assert qca.doi == doi
    assert qca.data == Decimal(ans)


def test_access_2f(contexts):
    constants, ans, _, _ = contexts
    assert constants.get("hARTREE energy in eV") == float(ans)


def test_access_2g(contexts):
    constants, ans, _, _ = contexts
    ref = {"label": "Hartree energy in eV", "units": "eV", "data": Decimal(ans)}
    dqca = constants.get("Hartree energy in eV", return_tuple=True).dict()

    for itm in ref:
        assert ref[itm] == dqca[itm]


def test_c_header():
    from qcelemental.physical_constants.context import write_c_header

    write_c_header("CODATA2018", "header.h")
    os.remove("header.h")


@pytest.mark.xfail(True, reason="comparison data not available for installed repository", run=True, strict=False)
def test_constants_comparison():
    from qcelemental.physical_constants.context import run_comparison

    run_comparison("CODATA2014")


def test_representation():
    qcelemental.constants.string_representation()


def test_str():
    assert "PhysicalConstantsContext(" in str(qcelemental.constants)


def test_codata2022():
    with pytest.raises(KeyError) as e:
        qcelemental.PhysicalConstantsContext("CODATA2022")

    assert "only contexts {'CODATA2014', 'CODATA2018', } are currently supported" in str(e.value)


def test_codata_comparison():
    from qcelemental.physical_constants.context import run_internal_comparison

    run_internal_comparison("CODATA2014", "CODATA2018")



"""Functions to enumerate all perfect and maximum matchings in bipartite graph.

Implemented following the algorithms in the paper "Algorithms for Enumerating
All Perfect, Maximum and Maximal Matchings in Bipartite Graphs" by Takeaki Uno,
using numpy and networkx modules of python.

NOTICE: optimization needed.

Author: guangzhi XU (xugzhi1987@gmail.com; guangzhi.xu@outlook.com)
Update time: 2017-06-29 14:41:56.

Copied Dec 2017 LAB Adapted from https://github.com/Xunius/bipartite_matching
Updated Dec 2017 LAB for pep8, py3, more tests, starter_match, and simpler interface

"""
import numpy as np

# commented as untested [Apr 2019]
# def _plotGraph(graph):
#    """Plot graph using nodes as position number."""
#    import networkx as nx
#
#    import matplotlib.pyplot as plt
#    fig = plt.figure()
#    ax = fig.add_subplot(111)
#
#    pos = [(ii[1], ii[0]) for ii in graph.nodes()]
#    pos_dict = dict(zip(graph.nodes(), pos))
#    nx.draw(graph, pos=pos_dict, ax=ax, with_labels=True)
#    plt.show(block=False)
#    return


def _formDirected(g, match):
    """Form directed graph D from G and matching M.

    Parameters
    ----------
    g :
        Undirected bipartite graph. Nodes are separated by their
        'bipartite' attribute.
    match :
        List of edges forming a matching of `g`.

    Returns
    -------
    networkx.DiGraph
        Directed graph, with edges in `match` pointing from set-0
        (bipartite attribute==0) to set-1 (bipartite attrbiute==1), and
        the other edges in `g` but not in `match` pointing from set-1 to
        set-0.

    """
    import networkx as nx

    d = nx.DiGraph()

    for ee in g.edges():
        if ee in match or (ee[1], ee[0]) in match:
            if g.nodes[ee[0]]["bipartite"] == 0:
                d.add_edge(ee[0], ee[1])
            else:
                d.add_edge(ee[1], ee[0])
        else:
            if g.nodes[ee[0]]["bipartite"] == 0:
                d.add_edge(ee[1], ee[0])
            else:
                d.add_edge(ee[0], ee[1])

    return d


def _enumMaximumMatching(g, starter_match=None):
    """Find all maximum matchings in an undirected bipartite graph `g`.

    Parameters
    ----------
    g : networkx.Graph
        Undirected bipartite graph. Nodes are separated by their
        'bipartite' attribute.
    starter_match : dict, optional
        Single perfect match to inaugurate Uno's algorithm.

    Returns
    -------
    list
        Each is a list of edges forming a maximum matching of `g`.

    Author
    ------
    guangzhi XU (xugzhi1987@gmail.com; guangzhi.xu@outlook.com)
    Update time: 2017-05-21 20:04:51.

    """
    import networkx as nx

    all_matches = []

    # ----------------Find one matching M----------------
    if starter_match is None:
        match = nx.bipartite.hopcroft_karp_matching(g)
    else:
        match = starter_match

    # ---------------Re-orient match arcs---------------
    match2 = []
    for kk, vv in match.items():
        if g.nodes[kk]["bipartite"] == 0:
            match2.append((kk, vv))
    match = match2
    all_matches.append(match)

    # -----------------Enter recursion-----------------
    all_matches = _enumMaximumMatchingIter(g, match, all_matches, None)

    return all_matches


def _enumMaximumMatchingIter(g, match, all_matches, add_e=None):
    """Recurively search maximum matchings.

    Parameters
    ----------
    g :
        Undirected bipartite graph. Nodes are separated by their
        'bipartite' attribute.
    match :
        List of edges forming one maximum matching of `g`.
    all_matches :
        List, each is a list of edges forming a maximum matching of `g`.
        Newly found matchings will be appended into this list.
    add_e : tuple, optional
        Edge used to form subproblems. If not `None`, will be added to each
        newly found matchings.

    Returns
    -------
    list
        Updated list of all maximum matchings.

    Author
    ------
    guangzhi XU (xugzhi1987@gmail.com; guangzhi.xu@outlook.com)
    Update time: 2017-05-21 20:09:06.

    """
    import networkx as nx

    # ---------------Form directed graph D---------------
    d = _formDirected(g, match)

    # -----------------Find cycles in D-----------------
    cycles = list(nx.simple_cycles(d))

    if len(cycles) == 0:

        # ---------If no cycle, find a feasible path---------
        all_uncovered = set(g.nodes).difference(set([ii[0] for ii in match]))
        all_uncovered = all_uncovered.difference(set([ii[1] for ii in match]))
        all_uncovered = list(all_uncovered)

        # --------------If no path, terminiate--------------
        if len(all_uncovered) == 0:
            return all_matches

        # ----------Find a length 2 feasible path----------
        idx = 0
        uncovered = all_uncovered[idx]
        while True:

            if uncovered not in nx.isolates(g):
                paths = nx.single_source_shortest_path(d, uncovered, cutoff=2)
                len2paths = [vv for kk, vv in paths.items() if len(vv) == 3]

                if len(len2paths) > 0:
                    reversed = False
                    break

                # ----------------Try reversed path----------------
                paths_rev = nx.single_source_shortest_path(d.reverse(), uncovered, cutoff=2)
                len2paths = [vv for kk, vv in paths_rev.items() if len(vv) == 3]

                if len(len2paths) > 0:
                    reversed = True
                    break

            idx += 1
            if idx > len(all_uncovered) - 1:
                return all_matches

            uncovered = all_uncovered[idx]

        # -------------Create a new matching M'-------------
        len2path = len2paths[0]
        if reversed:
            len2path = len2path[::-1]
        len2path = list(zip(len2path[:-1], len2path[1:]))

        new_match = []
        for ee in d.edges():
            if ee in len2path:
                if g.nodes[ee[1]]["bipartite"] == 0:
                    new_match.append((ee[1], ee[0]))
            else:
                if g.nodes[ee[0]]["bipartite"] == 0:
                    new_match.append(ee)

        if add_e is not None:
            for ii in add_e:
                new_match.append(ii)

        all_matches.append(new_match)

        # ---------------------Select e---------------------
        e = set(len2path).difference(set(match))
        e = list(e)[0]

        # -----------------Form subproblems-----------------
        g_plus = g.copy()
        g_minus = g.copy()
        g_plus.remove_node(e[0])
        g_plus.remove_node(e[1])

        g_minus.remove_edge(e[0], e[1])

        add_e_new = [e]
        if add_e is not None:
            add_e_new.extend(add_e)

        all_matches = _enumMaximumMatchingIter(g_minus, match, all_matches, add_e)
        all_matches = _enumMaximumMatchingIter(g_plus, new_match, all_matches, add_e_new)

    else:
        # ----------------Find a cycle in D----------------
        cycle = cycles[0]
        cycle.append(cycle[0])
        cycle = list(zip(cycle[:-1], cycle[1:]))

        # -------------Create a new matching M'-------------
        new_match = []
        for ee in d.edges():
            if ee in cycle:
                if g.nodes[ee[1]]["bipartite"] == 0:
                    new_match.append((ee[1], ee[0]))
            else:
                if g.nodes[ee[0]]["bipartite"] == 0:
                    new_match.append(ee)

        if add_e is not None:
            for ii in add_e:
                new_match.append(ii)

        all_matches.append(new_match)

        # -----------------Choose an edge E-----------------
        e = set(match).intersection(set(cycle))
        e = list(e)[0]

        # -----------------Form subproblems-----------------
        g_plus = g.copy()
        g_minus = g.copy()
        g_plus.remove_node(e[0])
        g_plus.remove_node(e[1])
        g_minus.remove_edge(e[0], e[1])

        add_e_new = [e]
        if add_e is not None:
            add_e_new.extend(add_e)

        all_matches = _enumMaximumMatchingIter(g_minus, new_match, all_matches, add_e)
        all_matches = _enumMaximumMatchingIter(g_plus, match, all_matches, add_e_new)

    return all_matches


def _enumMaximumMatching2(g):
    """Find all maximum matchings in an undirected bipartite graph `g`.
    Similar to _enumMaximumMatching but implemented using adjacency matrix
    of graph for slight speed boost.

    Parameters
    ----------
    g:
        Undirected bipartite graph. Nodes are separated by their
        'bipartite' attribute.

    Returns
    -------
    list
        Each is a list of edges forming a maximum matching of `g`.

    Author
    ------
    guangzhi XU (xugzhi1987@gmail.com; guangzhi.xu@outlook.com)
    Update time: 2017-05-21 20:04:51.

    """
    import networkx as nx
    from scipy import sparse

    s1 = set(n for n, d in g.nodes(data=True) if d["bipartite"] == 0)
    s2 = set(g) - s1
    n1 = len(s1)
    nodes = list(s1) + list(s2)

    adj = nx.adjacency_matrix(g, nodes).tolil()
    all_matches = []

    # ----------------Find one matching----------------
    match = nx.bipartite.hopcroft_karp_matching(g)

    matchadj = np.zeros(adj.shape).astype("int")
    for kk, vv in match.items():
        matchadj[nodes.index(kk), nodes.index(vv)] = 1
    matchadj = sparse.lil_matrix(matchadj)

    all_matches.append(matchadj)

    # -----------------Enter recursion-----------------
    all_matches = _enumMaximumMatchingIter2(adj, matchadj, all_matches, n1, None, True)

    # ---------------Re-orient match arcs---------------
    all_matches2 = []
    for ii in all_matches:
        match_list = sparse.find(ii[:n1] == 1)
        m1 = [nodes[jj] for jj in match_list[0]]
        m2 = [nodes[jj] for jj in match_list[1]]
        match_list = zip(m1, m2)

        all_matches2.append(match_list)

    print("got all")
    return all_matches2


def _enumMaximumMatchingIter2(adj, matchadj, all_matches, n1, add_e=None, check_cycle=True):
    """Recurively search maximum matchings.
       Similar to _enumMaximumMatching but implemented using adjacency matrix
       of graph for a slight speed boost.

    Parameters
    ----------
    # g :
    #     Undirected bipartite graph. Nodes are separated by their
    #     'bipartite' attribute.
    # match :
    #     List of edges forming one maximum matching of `g`.
    # all_matches :
    #     List, each is a list of edges forming a maximum matching of `g`.
    #     Newly found matchings will be appended into this list.
    add_e : tuple, optional
        Edge used to form subproblems. If not `None`, will be added to each
        newly found matchings.

    Returns
    -------
    list
        Updated list of all maximum matchings.

    Author
    ------
    guangzhi XU (xugzhi1987@gmail.com; guangzhi.xu@outlook.com)
    Update time: 2017-05-21 20:09:06.

    """
    import networkx as nx
    from scipy import sparse

    # -------------------Find cycles-------------------
    if check_cycle:
        d = matchadj.multiply(adj)
        d[n1:, :] = adj[n1:, :] - matchadj[n1:, :].multiply(adj[n1:, :])

        dg = nx.from_numpy_matrix(d.toarray(), create_using=nx.DiGraph())
        cycles = list(nx.simple_cycles(dg))
        if len(cycles) == 0:
            check_cycle = False
        else:
            check_cycle = True

    if check_cycle:
        cycle = cycles[0]
        cycle.append(cycle[0])
        cycle = zip(cycle[:-1], cycle[1:])

        # --------------Create a new matching--------------
        new_match = matchadj.copy()
        for ee in cycle:
            if matchadj[ee[0], ee[1]] == 1:
                new_match[ee[0], ee[1]] = 0
                new_match[ee[1], ee[0]] = 0
                e = ee
            else:
                new_match[ee[0], ee[1]] = 1
                new_match[ee[1], ee[0]] = 1

        if add_e is not None:
            for ii in add_e:
                new_match[ii[0], ii[1]] = 1

        all_matches.append(new_match)

        # -----------------Form subproblems-----------------
        g_plus = adj.copy()
        g_minus = adj.copy()
        g_plus[e[0], :] = 0
        g_plus[:, e[1]] = 0
        g_plus[:, e[0]] = 0
        g_plus[e[1], :] = 0
        g_minus[e[0], e[1]] = 0
        g_minus[e[1], e[0]] = 0

        add_e_new = [e]
        if add_e is not None:
            add_e_new.extend(add_e)

        all_matches = _enumMaximumMatchingIter2(g_minus, new_match, all_matches, n1, add_e, check_cycle)
        all_matches = _enumMaximumMatchingIter2(g_plus, matchadj, all_matches, n1, add_e_new, check_cycle)

    else:
        # ---------------Find uncovered nodes---------------
        uncovered = np.where(np.sum(matchadj, axis=1) == 0)[0]

        if len(uncovered) == 0:
            return all_matches

        # ---------------Find feasible paths---------------
        paths = []
        for ii in uncovered:
            aa = adj[ii, :].dot(matchadj)
            if aa.sum() == 0:
                continue
            paths.append((ii, int(sparse.find(aa == 1)[1][0])))
            if len(paths) > 0:
                break

        if len(paths) == 0:
            return all_matches

        # ----------------------Find e----------------------
        feas1, feas2 = paths[0]
        e = (feas1, int(sparse.find(matchadj[:, feas2] == 1)[0]))

        # ----------------Create a new match----------------
        new_match = matchadj.copy()
        new_match[feas2, :] = 0
        new_match[:, feas2] = 0
        new_match[feas1, e[1]] = 1
        new_match[e[1], feas1] = 1

        if add_e is not None:
            for ii in add_e:
                new_match[ii[0], ii[1]] = 1

        all_matches.append(new_match)

        # -----------------Form subproblems-----------------
        g_plus = adj.copy()
        g_minus = adj.copy()
        g_plus[e[0], :] = 0
        g_plus[:, e[1]] = 0
        g_plus[:, e[0]] = 0
        g_plus[e[1], :] = 0
        g_minus[e[0], e[1]] = 0
        g_minus[e[1], e[0]] = 0

        add_e_new = [e]
        if add_e is not None:
            add_e_new.extend(add_e)

        all_matches = _enumMaximumMatchingIter2(g_minus, matchadj, all_matches, n1, add_e, check_cycle)
        all_matches = _enumMaximumMatchingIter2(g_plus, new_match, all_matches, n1, add_e_new, check_cycle)

    # if len(all_matches) % 1000 == 0:
    #    print('len', len(all_matches))

    # print('another')
    return all_matches


# commented as unused [Apr 2019]
# def _findCycle(adj, n1):
#    from scipy import sparse
#
#    path = []
#    visited = set()
#
#    def visit(v):
#        if v in visited:
#            return False
#        visited.add(v)
#        path.append(v)
#        neighbours = sparse.find(adj[v, :] == 1)[1]
#        for nn in neighbours:
#            if nn in path or visit(nn):
#                return True
#        path.remove(v)
#        return False
#
#    nodes = range(n1)
#    result = any(visit(v) for v in nodes)
#    return result, path


def uno(edges, match=None, verbose=1):
    """Perform Uno's algorithm to enumerate all equivalent matchings among
    the bipartite graph defined by `edges`. Optionally given a single
    known `match`.

    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.107.8179&rep=rep1&type=pdf

    "Algorithms for Enumerating All Perfect, Maximum and Maximal Matchings
    in Bipartite Graphs" by Takeaki UNO

    """
    import networkx as nx

    if match is None:
        p_match = None
    else:
        p_match = {(1, m[0]): (0, m[1]) for m in match}
        p_m_inv = {(0, m[1]): (1, m[0]) for m in match}
        p_match.update(p_m_inv)

    p_edges = [[(1, e[0]), (0, e[1])] for e in edges]
    if verbose >= 2:
        print("Edges:")
        for e in p_edges:
            print("\t", e)

    g = nx.Graph()
    for e in p_edges:
        g.add_node(e[0], bipartite=0)
        g.add_node(e[1], bipartite=1)
        if verbose >= 2:
            print("  Node:", e[0], e[1])
    g.add_edges_from(p_edges)

    all_matches = _enumMaximumMatching(g, starter_match=p_match)
    p_all_matches = [sorted([(pt[0][1], pt[1][1]) for pt in am]) for am in all_matches]

    return p_all_matches



from typing import Optional

import numpy as np
from pydantic import Field, validator

from ..util import blockwise_contract, blockwise_expand
from .basemodels import ProtoModel
from .types import Array

__all__ = ["AlignmentMill"]


class AlignmentMill(ProtoModel):
    """Facilitates the application of the simple transformation operations
    defined by ``shift``, ``rotation``, ``atommap`` arrays and ``mirror``
    boolean as recipe to the data structures
    describing Cartesian molecular coordinates. Attaches functions to
    transform the geometry, element list, gradient, etc. to the
    AlignmentRecipe. When `mirror` attribute (defaults to False) active,
    then molecular system can be substantively changed by procedure.

    """

    shift: Optional[Array[float]] = Field(None, description="Translation array (3,) for coordinates.")  # type: ignore
    rotation: Optional[Array[float]] = Field(None, description="Rotation array (3, 3) for coordinates.")  # type: ignore
    atommap: Optional[Array[int]] = Field(None, description="Atom exchange map (nat,) for coordinates.")  # type: ignore
    mirror: bool = Field(False, description="Do mirror invert coordinates?")

    class Config:
        force_skip_defaults = True

    @validator("shift")
    def _must_be_3(cls, v, values, **kwargs):
        try:
            v = v.reshape(3)
        except (ValueError, AttributeError):
            raise ValueError("Shift must be castable to shape (3,)!")
        return v

    @validator("rotation")
    def _must_be_33(cls, v, values, **kwargs):
        try:
            v = v.reshape(3, 3)
        except (ValueError, AttributeError):
            raise ValueError("Rotation must be castable to shape (3, 3)!")
        return v

    ### Non-Pydantic API functions

    def pretty_print(self, label: str = "") -> str:
        width = 40
        text = []
        text.append("-" * width)
        text.append("{:^{width}}".format("AlignmentMill", width=width))
        if label:
            text.append("{:^{width}}".format(label, width=width))
        text.append("-" * width)
        text.append("Mirror:   {}".format(self.mirror))
        text.append("Atom Map: {}".format(self.atommap))
        text.append("Shift:    {}".format(self.shift))
        text.append("Rotation:")
        text.append("{}".format(self.rotation))
        text.append("-" * width)
        return "\n".join(x.rstrip() for x in text)

    def align_coordinates(self, geom, *, reverse=False) -> Array:
        """suitable for geometry or displaced geometry"""

        algeom = np.copy(geom)
        if reverse:
            algeom = algeom.dot(self.rotation)
            algeom = algeom + self.shift
            if self.mirror:
                algeom[:, 1] *= -1.0
        else:
            if self.mirror:
                algeom[:, 1] *= -1.0
            algeom = algeom - self.shift
            algeom = algeom.dot(self.rotation)
        algeom = algeom[self.atommap, :]
        # mirror-wise, rsm/msr == rms/msr

        return algeom

    def align_atoms(self, ats):
        """suitable for masses, symbols, Zs, etc."""

        return ats[self.atommap]

    def align_vector(self, vec):
        """suitable for vector attached to molecule"""

        # sensible? TODO
        # alvec = np.copy(vec)
        # if self.mirror:
        #    alvec[:, 1] *= -1
        return vec.dot(self.rotation)

    def align_gradient(self, grad) -> Array:
        """suitable for vector system attached to atoms"""

        # sensible? TODO
        # algrad = np.copy(grad)
        # if self.mirror:
        #    algrad[:, 1] *= -1
        algrad = grad.dot(self.rotation)
        algrad = algrad[self.atommap]

        return algrad

    def align_hessian(self, hess) -> Array:
        blocked_hess = blockwise_expand(hess, (3, 3), False)
        alhess = np.zeros_like(blocked_hess)

        nat = blocked_hess.shape[0]
        for iat in range(nat):
            for jat in range(nat):
                alhess[iat, jat] = (self.rotation.T).dot(blocked_hess[iat, jat].dot(self.rotation))

        alhess = alhess[np.ix_(self.atommap, self.atommap)]

        alhess = blockwise_contract(alhess)
        return alhess

    def align_vector_gradient(self, mu_derivatives):
        """Align the nuclear gradients of vector components (e.g. dipole derivatives)."""
        # Input data is assumed to be organized into outermost x, y, z vector components.
        # Organize derivatives for each atom into 3x3 and transform it.
        mu_x, mu_y, mu_z = mu_derivatives
        nat = mu_x.shape[0] // 3
        al_mu = np.zeros((3, 3 * nat))

        Datom = np.zeros((3, 3))  # atom whose nuclear derivatives are taken
        for at in range(nat):
            Datom.fill(0)
            Datom[0, :] = mu_x[3 * self.atommap[at] : 3 * self.atommap[at] + 3]
            Datom[1, :] = mu_y[3 * self.atommap[at] : 3 * self.atommap[at] + 3]
            Datom[2, :] = mu_z[3 * self.atommap[at] : 3 * self.atommap[at] + 3]
            Datom[:] = np.dot(self.rotation.T, np.dot(Datom, self.rotation))
            al_mu[0, 3 * at : 3 * at + 3] = Datom[0, :]
            al_mu[1, 3 * at : 3 * at + 3] = Datom[1, :]
            al_mu[2, 3 * at : 3 * at + 3] = Datom[2, :]
        return al_mu

    def align_system(self, geom, mass, elem, elez, uniq, *, reverse: bool = False):
        """For AlignmentRecipe `ar`, apply its translation, rotation, and atom map."""

        nugeom = self.align_coordinates(geom, reverse=reverse)
        numass = self.align_atoms(mass)
        nuelem = self.align_atoms(elem)
        nuelez = self.align_atoms(elez)
        nuuniq = self.align_atoms(uniq)

        return nugeom, numass, nuelem, nuelez, nuuniq

    def align_mini_system(self, geom, uniq, *, reverse: bool = False):
        """For AlignmentRecipe `ar`, apply its translation, rotation, and atom map."""

        nugeom = self.align_coordinates(geom, reverse=reverse)
        nuuniq = self.align_atoms(uniq)

        return nugeom, nuuniq



import copy

import numpy as np
import pytest

import qcelemental
from qcelemental.testing import compare_molrecs, tnm

from .addons import using_web, xfail_on_pubchem_busy

_string_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_string"}

subject4 = """pubchem:benzene"""

ans4 = {
    # fmt: off
    'geom': [-1.213100, -0.688400, 0.000000, -1.202800, 0.706400, 0.000100,
             -0.010300, -1.394800, 0.000000, 0.010400, 1.394800, -0.000100, 1.202800,
             -0.706300, 0.000000, 1.213100, 0.688400, 0.000000, -2.157700, -1.224400,
             0.000000, -2.139300, 1.256400, 0.000100, -0.018400, -2.480900, -0.000100,
             0.018400, 2.480800, 0.000000, 2.139400, -1.256300, 0.000100, 2.157700,
             1.224500, 0.000000
            ],
    # fmt: on
    "elbl": ["C", "C", "C", "C", "C", "C", "H", "H", "H", "H", "H", "H"],
    "units": "Angstrom",
    "molecular_charge": 0.0,
    "fragment_separators": [],
    "fragment_charges": [None],
    "fragment_multiplicities": [None],
    "fragment_files": [],
    "geom_hints": [],
    "hint_types": [],
    "name": "IUPAC benzene",
}

fullans4 = {
    # fmt: off
    'geom': np.array(
        [ -1.213100, -0.688400, 0.000000, -1.202800, 0.706400, 0.000100,
          -0.010300, -1.394800, 0.000000, 0.010400, 1.394800, -0.000100, 1.202800,
          -0.706300, 0.000000, 1.213100, 0.688400, 0.000000, -2.157700, -1.224400,
          0.000000, -2.139300, 1.256400, 0.000100, -0.018400, -2.480900, -0.000100,
          0.018400, 2.480800, 0.000000, 2.139400, -1.256300, 0.000100, 2.157700,
          1.224500, 0.000000,
        ]
    ),
    # fmt: on
    "elea": np.array([12, 12, 12, 12, 12, 12, 1, 1, 1, 1, 1, 1]),
    "elez": np.array([6, 6, 6, 6, 6, 6, 1, 1, 1, 1, 1, 1]),
    "elem": np.array(["C", "C", "C", "C", "C", "C", "H", "H", "H", "H", "H", "H"]),
    "mass": np.array(
        [12.0, 12.0, 12.0, 12.0, 12.0, 12.0, 1.00782503, 1.00782503, 1.00782503, 1.00782503, 1.00782503, 1.00782503]
    ),
    "real": np.array([True, True, True, True, True, True, True, True, True, True, True, True]),
    "elbl": np.array(["", "", "", "", "", "", "", "", "", "", "", ""]),
    "units": "Angstrom",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "name": "IUPAC benzene",
}


@using_web
def test_pubchem_4a():
    subject = subject4
    fullans = copy.deepcopy(fullans4)
    fullans["provenance"] = _string_prov_stamp

    with xfail_on_pubchem_busy():
        final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_molrecs(ans4, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


@using_web
def test_pubchem_4b():
    """user units potentially contradicting pubchem units"""
    subject = subject4 + "\nunits au"

    with pytest.raises(qcelemental.MoleculeFormatError):
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject, return_processed=True)


@using_web
def test_pubchem_4c():
    subject = """
pubchem  : 241
"""
    ans = copy.deepcopy(ans4)
    ans["name"] = "benzene"
    fullans = copy.deepcopy(fullans4)
    fullans["name"] = "benzene"
    fullans["provenance"] = _string_prov_stamp

    with xfail_on_pubchem_busy():
        final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, name="benzene", verbose=2)
    assert compare_molrecs(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_pubchem_error_d():
    subject = """
    pubchem:-55
"""

    with pytest.raises(qcelemental.ValidationError):
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject, return_processed=True)


def test_pubchem_error_e():
    # no 3D structure available
    subject = """
pubchem : sodium benzenesulfonate

"""

    with pytest.raises(qcelemental.ValidationError):
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject)


def test_pubchem_error_f():
    subject = """
    pubchem: 100000000000000
"""

    with pytest.raises(qcelemental.ValidationError):
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject, return_processed=True)


@using_web
def test_pubchem_multiout_g():
    subject = """
    #pubchem: gobbledegook
    #pubchem: c6h12o
    #pubchem: formaldehyde*
    pubchem: tropolone*
"""

    with pytest.raises(qcelemental.ChoicesError) as e:
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject, return_processed=True)

    try:
        with xfail_on_pubchem_busy():
            qcelemental.molparse.from_string(subject, return_processed=True)
    except qcelemental.ChoicesError as e:
        assert e.choices[10789] == "2-hydroxycyclohepta-2,4,6-trien-1-one"
        assert e.choices[193687] == "2-hydroxy-3-iodo-6-propan-2-ylcyclohepta-2,4,6-trien-1-one"


subject13 = """pubchem :ammonium\n"""

ans13 = {
    "name": "IUPAC azanium",
    "molecular_charge": 1.0,
    "units": "Angstrom",
    "fragment_files": [],
    "hint_types": [],
    "geom_hints": [],
    "elbl": ["N", "H", "H", "H", "H"],
    "fragment_separators": [],
    "fragment_charges": [None],
    "fragment_multiplicities": [None],
    # fmt: off
    'geom': [
        0.0, 0.0, 0.0, 0.9645, 0.2796, 0.2138, 0.0075, -0.886, -0.5188,
        -0.5235, -0.1202, 0.8751, -0.4486, 0.7266, -0.5701,
    ],
    # fmt: on
}

fullans13 = {
    "name": "IUPAC azanium",
    "units": "Angstrom",
    # fmt: off
    'geom': np.array([
        0.0, 0.0, 0.0, 0.9645, 0.2796, 0.2138, 0.0075, -0.886, -0.5188,
        -0.5235, -0.1202, 0.8751, -0.4486, 0.7266, -0.5701,
        ]),
    # fmt: on
    "elea": np.array([14, 1, 1, 1, 1]),
    "elez": np.array([7, 1, 1, 1, 1]),
    "elem": np.array(["N", "H", "H", "H", "H"]),
    "mass": np.array([14.003074, 1.00782503, 1.00782503, 1.00782503, 1.00782503]),
    "real": np.array([True, True, True, True, True]),
    "elbl": np.array(["", "", "", "", ""]),
    "fragment_separators": [],
    "molecular_charge": 1.0,
    "fragment_charges": [1.0],
    "molecular_multiplicity": 1,
    "fragment_multiplicities": [1],
    "fix_com": False,
    "fix_orientation": False,
}


@using_web
def test_pubchem_13h():
    subject = subject13
    fullans = copy.deepcopy(fullans13)
    fullans["provenance"] = _string_prov_stamp

    with xfail_on_pubchem_busy():
        final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_molrecs(ans13, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


@using_web
def test_pubchem_13i():
    subject = "PubChem:223"
    fullans = copy.deepcopy(fullans13)
    fullans["provenance"] = _string_prov_stamp

    with xfail_on_pubchem_busy():
        final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_molrecs(ans13, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")



from typing import Any, Dict

import numpy as np


class TypedArray(np.ndarray):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        try:
            v = np.asarray(v, dtype=cls._dtype)
        except ValueError:
            raise ValueError("Could not cast {} to NumPy Array!".format(v))

        return v

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        dt = cls._dtype
        if dt is int or np.issubdtype(dt, np.integer):
            items = {"type": "number", "multipleOf": 1.0}
        elif dt is float or np.issubdtype(dt, np.floating):
            items = {"type": "number"}
        elif dt is str or np.issubdtype(dt, np.string_):
            items = {"type": "string"}
        elif dt is bool or np.issubdtype(dt, np.bool_):
            items = {"type": "boolean"}
        field_schema.update(type="array", items=items)


class ArrayMeta(type):
    def __getitem__(self, dtype):
        return type("Array", (TypedArray,), {"_dtype": dtype})


class Array(np.ndarray, metaclass=ArrayMeta):
    pass



import math
import re
from typing import Dict, List

import numpy as np

from ..physical_constants import constants


def distance_matrix(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Euclidean distance matrix between rows of arrays `a` and `b`. Equivalent to
    `scipy.spatial.distance.cdist(a, b, 'euclidean')`. Returns a.shape[0] x b.shape[0] array.

    """
    assert a.shape[1] == b.shape[1], """Inner dimensions do not match"""
    distm = np.zeros([a.shape[0], b.shape[0]])
    for i in range(a.shape[0]):
        distm[i] = np.linalg.norm(a[i] - b, axis=1)
    return distm


def update_with_error(a: Dict, b: Dict, path=None) -> Dict:
    """Merges `b` into `a` like dict.update; however, raises KeyError if values of a
    key shared by `a` and `b` conflict.

    Adapted from: https://stackoverflow.com/a/7205107

    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                update_with_error(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif a[key] is None:
                a[key] = b[key]
            elif (
                isinstance(a[key], (list, tuple))
                and not isinstance(a[key], str)
                and isinstance(b[key], (list, tuple))
                and not isinstance(b[key], str)
                and len(a[key]) == len(b[key])
                and all((av is None or av == bv) for av, bv in zip(a[key], b[key]))
            ):  # yapf: disable
                a[key] = b[key]
            else:
                raise KeyError("Conflict at {}: {} vs. {}".format(".".join(path + [str(key)]), a[key], b[key]))
        else:
            a[key] = b[key]
    return a


def standardize_efp_angles_units(units: str, geom_hints: List[List[float]]) -> List[List[float]]:
    """Applies to the pre-validated xyzabc or points hints in `geom_hints`
    the libefp default (1) units of [a0] and (2) radian angle range of
    (-pi, pi]. The latter is handy since this is how libefp returns hints

    """

    def radrge(radang):
        """Adjust `radang` by 2pi into (-pi, pi] range."""
        if radang > math.pi:
            return radang - 2 * math.pi
        elif radang <= -math.pi:
            return radang + 2 * math.pi
        else:
            return radang

    if units == "Angstrom":
        iutau = 1.0 / constants.bohr2angstroms
    else:
        iutau = 1.0

    hints = []
    for hint in geom_hints:
        if len(hint) == 6:
            x, y, z = [i * iutau for i in hint[:3]]
            a, b, c = [radrge(i) for i in hint[3:]]
            hints.append([x, y, z, a, b, c])
        if len(hint) == 9:
            points = [i * iutau for i in hint]
            hints.append(points)

    return hints


def filter_comments(string: str) -> str:
    """Remove from `string` any Python-style comments ('#' to end of line)."""

    return re.sub(r"(^|[^\\])#.*", "", string)


def unnp(dicary: Dict, _path=None, *, flat: bool = False) -> Dict:
    """Return `dicary` with any ndarray values replaced by lists.

    Parameters
    ----------
    dicary: dict
        Dictionary where any internal iterables are dict or list.
    flat : bool, optional
        Whether the returned lists are flat or nested.

    Returns
    -------
    dict
        Input with any ndarray values replaced by lists.

    """

    if _path is None:
        _path = []

    ndicary: Dict = {}
    for k, v in dicary.items():
        if isinstance(v, dict):
            ndicary[k] = unnp(v, _path + [str(k)], flat=flat)
        elif isinstance(v, list):
            # relying on Py3.6+ ordered dict here
            fakedict = {kk: vv for kk, vv in enumerate(v)}
            tolisted = unnp(fakedict, _path + [str(k)], flat=flat)
            ndicary[k] = list(tolisted.values())
        else:
            try:
                v.shape
            except AttributeError:
                ndicary[k] = v
            else:
                if flat:
                    ndicary[k] = v.ravel().tolist()
                else:
                    ndicary[k] = v.tolist()
    return ndicary


def _norm(points) -> float:
    """
    Return the Frobenius norm across axis=-1, NumPy's internal norm is crazy slow (~4x)
    """

    tmp = np.atleast_2d(points)
    return np.sqrt(np.einsum("ij,ij->i", tmp, tmp))


def measure_coordinates(coordinates, measurements, degrees=False):
    """
    Measures a geometry array based on 0-based indices provided, automatically detects distance, angle,
    and dihedral based on length of measurement input.
    """

    coordinates = np.atleast_2d(coordinates)
    num_coords = coordinates.shape[0]

    single = False
    if isinstance(measurements[0], int):
        measurements = [measurements]
        single = True

    ret = []
    for num, m in enumerate(measurements):
        if any(x >= num_coords for x in m):
            raise ValueError(f"An index of measurement {num} is out of bounds.")

        kwargs = {}
        if len(m) == 2:
            func = compute_distance
        elif len(m) == 3:
            func = compute_angle
            kwargs = {"degrees": degrees}
        elif len(m) == 4:
            func = compute_dihedral
            kwargs = {"degrees": degrees}
        else:
            raise KeyError(f"Unrecognized number of arguments for measurement {num}, found {len(m)}, expected 2-4.")

        val = func(*[coordinates[x] for x in m], **kwargs)
        ret.append(float(val))

    if single:
        return ret[0]
    else:
        return ret


def compute_distance(points1, points2) -> np.ndarray:
    """
    Computes the distance between the provided points on a per-row basis.

    Parameters
    ----------
    points1 : array-like
        The first list of points, can be 1D or 2D
    points2 : array-like
        The second list of points, can be 1D or 2D

    Returns
    -------
    distances : np.ndarray
        The array of distances between points1 and points2

    Notes
    -----
    Units are not considered inside these expressions, please preconvert to the same units before using.

    See Also
    --------
    distance_matrix
        Computes the distance between the provided points in all rows.
        compute_distance result is the diagonal of the distance_matrix result.

    """
    points1 = np.atleast_2d(points1)
    points2 = np.atleast_2d(points2)

    return _norm(points1 - points2)


def compute_angle(points1, points2, points3, *, degrees: bool = False) -> np.ndarray:
    """
    Computes the angle (p1, p2 [vertex], p3) between the provided points on a per-row basis.

    Parameters
    ----------
    points1 : np.ndarray
        The first list of points, can be 1D or 2D
    points2 : np.ndarray
        The second list of points, can be 1D or 2D
    points3 : np.ndarray
        The third list of points, can be 1D or 2D
    degrees : bool, options
        Returns the angle in degrees rather than radians if True

    Returns
    -------
    angles : np.ndarray
        The angle between the three points in radians

    Notes
    -----
    Units are not considered inside these expressions, please preconvert to the same units before using.
    """

    points1 = np.atleast_2d(points1)
    points2 = np.atleast_2d(points2)
    points3 = np.atleast_2d(points3)

    v12 = points1 - points2
    v23 = points2 - points3

    denom = _norm(v12) * _norm(v23)
    cosine_angle = np.einsum("ij,ij->i", v12, v23) / denom

    angle = np.pi - np.arccos(cosine_angle)

    if degrees:
        return np.degrees(angle)
    else:
        return angle


def compute_dihedral(points1, points2, points3, points4, *, degrees: bool = False) -> np.ndarray:
    """
    Computes the dihedral angle (p1, p2, p3, p4) between the provided points on a per-row basis using the Praxeolitic formula.

    Parameters
    ----------
    points1 : np.ndarray
        The first list of points, can be 1D or 2D
    points2 : np.ndarray
        The second list of points, can be 1D or 2D
    points3 : np.ndarray
        The third list of points, can be 1D or 2D
    points4 : np.ndarray
        The third list of points, can be 1D or 2D
    degrees : bool, options
        Returns the dihedral angle in degrees rather than radians if True

    Returns
    -------
    dihedrals : np.ndarray
        The dihedral angle between the four points in radians

    Notes
    -----
    Units are not considered inside these expressions, please preconvert to the same units before using.
    """

    # FROM: https://stackoverflow.com/questions/20305272/

    points1 = np.atleast_2d(points1)
    points2 = np.atleast_2d(points2)
    points3 = np.atleast_2d(points3)
    points4 = np.atleast_2d(points4)

    # Build the three vectors
    v1 = -1.0 * (points2 - points1)
    v2 = points3 - points2
    v3 = points4 - points3

    # Normalize the central vector
    v2 = v2 / _norm(v2)

    # v = projection of b0 onto plane perpendicular to b1
    #   = b0 minus component that aligns with b1
    # w = projection of b2 onto plane perpendicular to b1
    #   = b2 minus component that aligns with b1
    v = v1 - np.einsum("ij,ij->i", v1, v1) * v2
    w = v3 - np.einsum("ij,ij->i", v3, v2) * v2

    # angle between v and w in a plane is the torsion angle
    # v and w may not be normalized but that's fine since tan is y/x
    x = np.einsum("ij,ij->i", v, w)
    y = np.einsum("ij,ij->i", np.cross(v2, v), w)
    angle = np.arctan2(y, x)

    if degrees:
        return np.degrees(angle)
    else:
        return angle



import pprint
import re
from copy import deepcopy

import numpy as np

from ..exceptions import ValidationError
from ..physical_constants import constants
from ..util import provenance_stamp, unnp, update_with_error
from .chgmult import validate_and_fill_chgmult
from .nucleus import reconcile_nucleus
from .regex import VERSION_PATTERN


def from_input_arrays(
    *,
    enable_qm=True,
    enable_efp=True,
    missing_enabled_return_qm="error",
    missing_enabled_return_efp="error",
    # qm
    geom=None,
    elea=None,
    elez=None,
    elem=None,
    mass=None,
    real=None,
    elbl=None,
    name=None,
    units="Angstrom",
    input_units_to_au=None,
    fix_com=None,
    fix_orientation=None,
    fix_symmetry=None,
    fragment_separators=None,
    fragment_charges=None,
    fragment_multiplicities=None,
    molecular_charge=None,
    molecular_multiplicity=None,
    # efp
    fragment_files=None,
    hint_types=None,
    geom_hints=None,
    # qm-vz
    geom_unsettled=None,
    variables=None,
    # processing details
    speclabel=True,
    tooclose=0.1,
    zero_ghost_fragments=False,
    nonphysical=False,
    mtol=1.0e-3,
    copy=True,
    verbose=1,
):
    """Compose a Molecule dict from unvalidated arrays and variables
    in multiple domains.

    Drives :py:func:`qcelemental.molparse.from_arrays` for sucessive
    domains and hooks them together (e.g., impose `fix_com` on "qm"
    when "efp" present.

    """
    molinit = {}
    if enable_qm:
        molinit["qm"] = {}
    if enable_efp:
        molinit["efp"] = {}

    if enable_efp:
        processed = from_arrays(
            domain="efp",
            missing_enabled_return=missing_enabled_return_efp,
            units=units,
            input_units_to_au=input_units_to_au,
            fix_com=fix_com,
            fix_orientation=fix_orientation,
            fix_symmetry=fix_symmetry,
            fragment_files=fragment_files,
            hint_types=hint_types,
            geom_hints=geom_hints,
            # which other processing details needed?
            verbose=verbose,
        )
        update_with_error(molinit, {"efp": processed})
        if molinit["efp"] == {}:
            del molinit["efp"]

    efp_present = enable_efp and "efp" in molinit and bool(len(molinit["efp"]["geom_hints"]))
    if efp_present:
        fix_com = True
        fix_orientation = True
        fix_symmetry = "c1"

    if enable_qm:
        dm = "qmvz" if geom_unsettled else "qm"
        processed = from_arrays(
            domain=dm,
            missing_enabled_return=missing_enabled_return_qm,
            geom=geom,
            elea=elea,
            elez=elez,
            elem=elem,
            mass=mass,
            real=real,
            elbl=elbl,
            name=name,
            units=units,
            input_units_to_au=input_units_to_au,
            fix_com=fix_com,
            fix_orientation=fix_orientation,
            fix_symmetry=fix_symmetry,
            fragment_separators=fragment_separators,
            fragment_charges=fragment_charges,
            fragment_multiplicities=fragment_multiplicities,
            molecular_charge=molecular_charge,
            molecular_multiplicity=molecular_multiplicity,
            geom_unsettled=geom_unsettled,
            variables=variables,
            # processing details
            speclabel=speclabel,
            tooclose=tooclose,
            zero_ghost_fragments=zero_ghost_fragments,
            nonphysical=nonphysical,
            mtol=mtol,
            copy=copy,
            verbose=1,
        )
        update_with_error(molinit, {"qm": processed})
        if molinit["qm"] == {}:
            del molinit["qm"]

    return molinit


def from_arrays(
    *,
    geom=None,
    elea=None,
    elez=None,
    elem=None,
    mass=None,
    real=None,
    elbl=None,
    name=None,
    units="Angstrom",
    input_units_to_au=None,
    fix_com=None,
    fix_orientation=None,
    fix_symmetry=None,
    fragment_separators=None,
    fragment_charges=None,
    fragment_multiplicities=None,
    molecular_charge=None,
    molecular_multiplicity=None,
    comment=None,
    provenance=None,
    connectivity=None,
    fragment_files=None,
    hint_types=None,
    geom_hints=None,
    geom_unsettled=None,
    variables=None,
    domain="qm",
    missing_enabled_return="error",
    np_out=True,
    speclabel=True,
    tooclose=0.1,
    zero_ghost_fragments=False,
    nonphysical=False,
    mtol=1.0e-3,
    copy=True,
    verbose=1,
):
    """Compose a Molecule dict from unvalidated arrays and variables, returning dict.

    See fields of Return molrec below. Required parameters (for QM XYZ)
    are `geom` and one of `elem`, `elez`, `elbl` (`speclabel=True`)

    Parameters
    ----------
    geom : array-like
        (nat, 3) or (3 * nat, ) ndarray or list o'lists of Cartesian coordinates.
    fragment_separators : array-like of int, optional
        (nfr - 1, ) list of atom indices at which to split `geom` into fragments.
    elbl : ndarray of str
        (nat, ) Label extending `elem` symbol, possibly conveying ghosting, isotope, mass, tagging information.
    tooclose : float, optional
        Interatom distance (native `geom` units) nearer than which atoms not allowed.
    nonphysical : bool, optional
        Do allow masses outside an element's natural range to pass validation?
    speclabel : bool, optional
        If `True`, interpret `elbl` as potentially full nucleus spec including
        ghosting, isotope, mass, tagging information, e.g., `@13C_mine` or
        `He4@4.01`. If `False`, interpret `elbl` as only the user/tagging
        extension to nucleus label, e.g. `_mine` or `4` in the previous examples.
    missing_enabled_return : {'minimal', 'none', 'error'}
        What to do when an enabled domain is of zero-length? Respectively, return
        a fully valid but empty molrec, return empty dictionary, or throw error.
    np_out : bool, optional
        When `True`, fields geom, elea, elez, elem, mass, real, elbl will be ndarray.
        Use `False` to get a json-able version.

    Returns
    -------
    molrec : dict
        Molecule dictionary spec follows. Its principles are

        (1) contents are fully validated and defaulted - no error
        checking necessary,

        (2) contents may be mildly redundant - atomic numbers and
        element symbols present,

        (3) big system, nat-length single-type arrays, not small system,
        nat-number heterogeneous objects,

        (4) some fields are optional (e.g., fix_symmetry) but largely
        self-describing so units or fix_com must be present.

        (5) apart from some mild optional fields, _all_ fields will
        be present (corollary of "fully validated and defaulted") - no
        need to check for every key. in some cases like efp, keys will
        appear in blocks, so pre-handshake there will be a few hint keys
        and post-handshake they will be joined by full qm-like molrec.

        (6) molrec should be idempotent through this function (equiv to
        schema validator) but are not idempotent throughout its life. if
        fields permit, frame may be changed. Future? if fields permit,
        mol may be symmetrized. Coordinates and angles may change units
        or range if program returns them in only one form.

    name : str, optional
        Label for molecule; should be valid Python identifier.
    units : {'Angstrom', 'Bohr'}
        Units for `geom`.
    input_units_to_au : float, optional
        If `units='Angstrom'`, overrides consumer's value for [A]-->[a0] conversion.
    fix_com : bool
        Whether translation of `geom` is allowed or disallowed.
    fix_orientation : bool
        Whether rotation of `geom` is allowed or disallowed.
    fix_symmetry : str, optional
        Maximal point group symmetry which `geom` should be treated. Lowercase.
    geom : ndarray of float
        (3 * nat, ) Cartesian coordinates in `units`.
    elea : ndarray of int
        (nat, ) Mass number for atoms, if known isotope, else -1.
    elez : ndarray of int
        (nat, ) Number of protons, nuclear charge for atoms.
    elem : ndarray of str
        (nat, ) Element symbol for atoms.
    mass : ndarray of float
        (nat, ) Atomic mass [u] for atoms.
    real : ndarray of bool
        (nat, ) Real/ghostedness for atoms.
    elbl : ndarray of str
        (nat, ) Label with any tagging information from element spec.
    fragment_separators : list of int
        (nfr - 1, ) list of atom indices at which to split `geom` into fragments.
    fragment_charges : list of float
        (nfr, ) list of charge allocated to each fragment.
    fragment_multiplicities : list of int
        (nfr, ) list of multiplicity allocated to each fragment.
    molecular_charge : float
        total charge on system.
    molecular_multiplicity : int
        total multiplicity on system.
    comment : str, optional
        Additional comment for molecule.
    provenance : dict of str
        Accumulated history of molecule, with fields "creator", "version", "routine".
    connectivity : list of tuples of int, optional
        (nbond, 3) list of (0-indexed) (atomA, atomB, bond_order) (int, int, double) tuples

    EFP extension (this + units is minimal)

    fragment_files : list of str
        (nfr, ) lowercased names of efp meat fragment files.
    hint_types : {'xyzabc', 'points'}
        (nfr, ) type of fragment orientation hint.
    geom_hints : list of lists of float
        (nfr, ) inner lists have length 6 (xyzabc; to orient the center) or
        9 (points; to orient the first three atoms) of the EFP fragment.

    QMVZ extension (geom_unsettled replaces geom)

    geom_unsettled : list of lists of str
        (nat, ) all-string Cartesian and/or zmat anchor and value contents
        mixing anchors, values, and variables.
    variables : list of pairs
        (nvar, 2) pairs of variables (str) and values (float). May be incomplete.

    Raises
    ------
    qcelemental.ValidationError
        For most anything wrong.

    """
    # <<  domain sorting  >>
    available_domains = ["qm", "efp", "qmvz"]
    if domain not in available_domains:
        raise ValidationError(
            "Topology domain {} not available for processing. Choose among {}".format(domain, available_domains)
        )

    if domain == "qm" and (geom is None or np.asarray(geom).size == 0):
        if missing_enabled_return == "none":
            return {}
        elif missing_enabled_return == "minimal":
            geom = []
        else:
            raise ValidationError("""For domain 'qm', `geom` must be provided.""")
    if domain == "efp" and (geom_hints is None or np.asarray(geom_hints).size == 0):
        if missing_enabled_return == "none":
            return {}
        elif missing_enabled_return == "minimal":
            geom_hints = []
            fragment_files = []
            hint_types = []
        else:
            raise ValidationError("""For domain 'efp', `geom_hints` must be provided.""")

    molinit = {}
    extern = False

    processed = validate_and_fill_units(
        name=name,
        units=units,
        input_units_to_au=input_units_to_au,
        comment=comment,
        provenance=provenance,
        connectivity=connectivity,
        always_return_iutau=False,
    )
    processed["provenance"] = provenance_stamp(__name__)
    update_with_error(molinit, processed)

    if domain == "efp":
        processed = validate_and_fill_efp(fragment_files=fragment_files, hint_types=hint_types, geom_hints=geom_hints)
        update_with_error(molinit, processed)
        extern = bool(len(molinit["geom_hints"]))

    if domain == "qm" or (domain == "efp" and geom is not None) or domain == "qmvz":
        if domain == "qmvz":
            processed = validate_and_fill_unsettled_geometry(geom_unsettled=geom_unsettled, variables=variables)
            update_with_error(molinit, processed)
            nat = len(molinit["geom_unsettled"])

        else:
            processed = validate_and_fill_geometry(geom=geom, tooclose=tooclose, copy=copy)
            update_with_error(molinit, processed)
            nat = molinit["geom"].shape[0] // 3

        processed = validate_and_fill_nuclei(
            nat,
            elea=elea,
            elez=elez,
            elem=elem,
            mass=mass,
            real=real,
            elbl=elbl,
            speclabel=speclabel,
            nonphysical=nonphysical,
            mtol=mtol,
            verbose=verbose,
        )
        update_with_error(molinit, processed)

        processed = validate_and_fill_fragments(
            nat,
            fragment_separators=fragment_separators,
            fragment_charges=fragment_charges,
            fragment_multiplicities=fragment_multiplicities,
        )
        update_with_error(molinit, processed)

        Z_available = molinit["elez"] * molinit["real"] * 1.0
        processed = validate_and_fill_chgmult(
            zeff=Z_available,
            fragment_separators=molinit["fragment_separators"],
            molecular_charge=molecular_charge,
            fragment_charges=molinit["fragment_charges"],
            molecular_multiplicity=molecular_multiplicity,
            fragment_multiplicities=molinit["fragment_multiplicities"],
            zero_ghost_fragments=zero_ghost_fragments,
            verbose=verbose,
        )
        del molinit["fragment_charges"]  # sometimes safe update is too picky about overwriting v_a_f_fragments values
        del molinit["fragment_multiplicities"]
        update_with_error(molinit, processed)

    extern = domain == "efp"

    processed = validate_and_fill_frame(
        extern=extern, fix_com=fix_com, fix_orientation=fix_orientation, fix_symmetry=fix_symmetry
    )
    update_with_error(molinit, processed)

    if verbose >= 2:
        print("RETURN FROM qcel.molparse.from_arrays(domain={})".format(domain.upper()))
        pprint.pprint(molinit)

    if not np_out:
        molinit = unnp(molinit)

    return molinit


def validate_and_fill_units(
    name=None,
    units="Angstrom",
    input_units_to_au=None,
    comment=None,
    provenance=None,
    connectivity=None,
    always_return_iutau=False,
):
    molinit = {}

    if name is not None:
        molinit["name"] = name

    if comment is not None:
        molinit["comment"] = comment

    def validate_provenance(dicary):
        expected_prov_keys = ["creator", "routine", "version"]
        try:
            prov_keys = sorted(dicary.keys())
        except AttributeError:
            raise ValidationError("Provenance entry is not dictionary: {}".format(dicary))

        if prov_keys == expected_prov_keys:
            if not isinstance(dicary["creator"], str):
                raise ValidationError(
                    """Provenance key 'creator' should be string of creating program's name: {}""".format(
                        dicary["creator"]
                    )
                )
            if not re.fullmatch(VERSION_PATTERN, dicary["version"], re.VERBOSE):
                raise ValidationError(
                    """Provenance key 'version' should be a valid PEP 440 string: {}""".format(dicary["version"])
                )
            if not isinstance(dicary["routine"], str):
                raise ValidationError(
                    """Provenance key 'routine' should be string of creating function's name: {}""".format(
                        dicary["routine"]
                    )
                )
            return True
        else:
            raise ValidationError("Provenance keys ({}) incorrect: {}".format(expected_prov_keys, prov_keys))

    if provenance is None:
        molinit["provenance"] = {}
    else:
        if validate_provenance(provenance):
            molinit["provenance"] = deepcopy(provenance)

    if connectivity is not None:
        conn = []
        try:
            for (at1, at2, bondorder) in connectivity:
                if not (float(at1)).is_integer() or at1 < 0:  # or at1 >= nat:
                    raise ValidationError("""Connectivity first atom should be int [0, nat): {}""".format(at1))
                if not (float(at2)).is_integer() or at2 < 0:  # or at2 >= nat:
                    raise ValidationError("""Connectivity second atom should be int [0, nat): {}""".format(at2))
                if bondorder < 0 or bondorder > 5:
                    raise ValidationError("""Connectivity bond order should be float [0, 5]: {}""".format(bondorder))
                conn.append((int(min(at1, at2)), int(max(at1, at2)), float(bondorder)))
            conn.sort(key=lambda tup: tup[0])
            molinit["connectivity"] = conn
        except ValueError:
            raise ValidationError(
                "Connectivity entry is not of form [(at1, at2, bondorder), ...]: {}".format(connectivity)
            )

    if units.capitalize() in ["Angstrom", "Bohr"]:
        molinit["units"] = units.capitalize()
    else:
        raise ValidationError("Invalid molecule geometry units: {}".format(units))

    if molinit["units"] == "Bohr":
        iutau = 1.0
    elif molinit["units"] == "Angstrom":
        iutau = 1.0 / constants.bohr2angstroms

    if input_units_to_au is not None:
        if abs(input_units_to_au - iutau) < 0.05:
            iutau = input_units_to_au
        else:
            raise ValidationError(
                """No big perturbations to physical constants! {} !~= {}""".format(iutau, input_units_to_au)
            )

    if always_return_iutau or input_units_to_au is not None:
        molinit["input_units_to_au"] = iutau

    return molinit


def validate_and_fill_frame(extern, fix_com=None, fix_orientation=None, fix_symmetry=None):

    if fix_com is True:
        com = True
    elif fix_com is False:
        if extern:
            raise ValidationError("Invalid fix_com ({}) with extern ({})".format(fix_com, extern))
        else:
            com = False
    elif fix_com is None:
        com = extern
    else:
        raise ValidationError("Invalid fix_com: {}".format(fix_com))

    if fix_orientation is True:
        orient = True
    elif fix_orientation is False:
        if extern:
            raise ValidationError("Invalid fix_orientation ({}) with extern ({})".format(fix_orientation, extern))
        else:
            orient = False
    elif fix_orientation is None:
        orient = extern
    else:
        raise ValidationError("Invalid fix_orientation: {}".format(fix_orientation))

    symm = None
    if extern:
        if fix_symmetry is None:
            symm = "c1"
        elif fix_symmetry.lower() == "c1":
            symm = "c1"
        else:
            raise ValidationError("Invalid (non-C1) fix_symmetry ({}) with extern ({})".format(fix_symmetry, extern))
    else:
        if fix_symmetry is not None:
            symm = fix_symmetry.lower()

    molinit = {}
    molinit["fix_com"] = com
    molinit["fix_orientation"] = orient
    if symm:
        molinit["fix_symmetry"] = symm

    return molinit


def validate_and_fill_efp(fragment_files=None, hint_types=None, geom_hints=None):

    if (
        fragment_files is None
        or hint_types is None
        or geom_hints is None
        or fragment_files == [None]
        or hint_types == [None]
        or geom_hints == [None]
        or not (len(fragment_files) == len(hint_types) == len(geom_hints))
    ):

        raise ValidationError(
            """Missing or inconsistent length among efp quantities: fragment_files ({}), hint_types ({}), and geom_hints ({})""".format(
                fragment_files, hint_types, geom_hints
            )
        )

    # NOTE: imposing case on file
    try:
        files = [f.lower() for f in fragment_files]
    except AttributeError:
        raise ValidationError("""fragment_files not strings: {}""".format(fragment_files))

    if all(f in ["xyzabc", "points", "rotmat"] for f in hint_types):
        types = hint_types
    else:
        raise ValidationError("""hint_types not among 'xyzabc', 'points', 'rotmat': {}""".format(hint_types))

    hints = []
    hlen = {"xyzabc": 6, "points": 9, "rotmat": 12}
    for ifr, fr in enumerate(geom_hints):
        try:
            hint = [float(f) for f in fr]
        except (ValueError, TypeError):
            raise ValidationError("""Un float-able elements in geom_hints[{}]: {}""".format(ifr, fr))

        htype = hint_types[ifr]
        if len(hint) == hlen[htype]:
            hints.append(hint)
        else:
            raise ValidationError("""EFP hint type {} not {} elements: {}""".format(htype, hlen[htype], hint))

    return {"fragment_files": files, "hint_types": types, "geom_hints": hints}


def validate_and_fill_geometry(geom=None, tooclose=0.1, copy=True):
    """Check `geom` for overlapping atoms. Return flattened"""

    npgeom = np.array(geom, copy=copy, dtype=np.float).reshape((-1, 3))

    # Upper triangular
    metric = tooclose ** 2
    tooclose_inds = []
    for x in range(npgeom.shape[0]):
        diffs = npgeom[x] - npgeom[x + 1 :]
        dists = np.einsum("ij,ij->i", diffs, diffs)

        # Record issues
        if np.any(dists < metric):
            indices = np.where(dists < metric)[0]
            tooclose_inds.extend([(x, y, dist) for y, dist in zip(indices + x + 1, dists[indices] ** 0.5)])

    if tooclose_inds:
        raise ValidationError(
            """Following atoms are too close: {}""".format([(i, j, dist) for i, j, dist in tooclose_inds])
        )

    return {"geom": npgeom.reshape((-1))}


def validate_and_fill_nuclei(
    nat,
    elea=None,
    elez=None,
    elem=None,
    mass=None,
    real=None,
    elbl=None,
    # processing details
    speclabel=True,
    nonphysical=False,
    mtol=1.0e-3,
    verbose=1,
):
    """Check the nuclear identity arrays for consistency and fill in knowable values."""

    if elea is None:
        elea = np.asarray([None] * nat)
    else:
        # -1 equivalent to None
        elea = np.asarray(elea)
        if -1 in elea:
            elea = np.array([(None if at == -1 else at) for at in elea])  # Rebuild to change dtype if needed.

    if elez is None:
        elez = np.asarray([None] * nat)
    else:
        elez = np.asarray(elez)

    if elem is None:
        elem = np.asarray([None] * nat)
    else:
        elem = np.asarray(elem)

    if mass is None:
        mass = np.asarray([None] * nat)
    else:
        mass = np.asarray(mass)

    if real is None:
        real = np.asarray([None] * nat)
    else:
        real = np.asarray(real)

    if elbl is None:
        elbl = np.asarray([None] * nat)
    else:
        elbl = np.asarray(elbl)

    if not ((nat,) == elea.shape == elez.shape == elem.shape == mass.shape == real.shape == elbl.shape):
        raise ValidationError(
            """Dimension mismatch natom {} among A {}, Z {}, E {}, mass {}, real {}, and elbl {}""".format(
                (nat,), elea.shape, elez.shape, elem.shape, mass.shape, real.shape, elbl.shape
            )
        )

    if nat:
        A, Z, E, mass, real, label = zip(
            *[
                reconcile_nucleus(
                    A=elea[at],
                    Z=elez[at],
                    E=elem[at],
                    mass=mass[at],
                    real=real[at],
                    label=elbl[at],
                    speclabel=speclabel,
                    nonphysical=nonphysical,
                    mtol=mtol,
                    verbose=verbose,
                )
                for at in range(nat)
            ]
        )
    else:
        A = Z = E = mass = real = label = []
    return {
        "elea": np.array(A, dtype=np.int),
        "elez": np.array(Z, dtype=np.int),
        "elem": np.array(E),
        "mass": np.array(mass, dtype=np.float),
        "real": np.array(real, dtype=np.bool),
        "elbl": np.array(label),
    }


def validate_and_fill_fragments(nat, fragment_separators=None, fragment_charges=None, fragment_multiplicities=None):
    """Check consistency of fragment specifiers wrt type and length. For
    charge & multiplicity, scientific defaults are not computed or applied;
    rather, missing slots are filled with `None` for later processing.

    """
    if fragment_separators is None:
        if fragment_charges is None and fragment_multiplicities is None:
            frs = []  # np.array([], dtype=np.int)  # if empty, needs to be both ndarray and int
            frc = [None]
            frm = [None]
        else:
            raise ValidationError(
                """Fragment quantities given without separation info: sep ({}), chg ({}), and mult ({})""".format(
                    fragment_separators, fragment_charges, fragment_multiplicities
                )
            )
    else:
        trial_geom = np.zeros((nat, 3))
        try:
            split_geom = np.split(trial_geom, fragment_separators, axis=0)
        except TypeError:
            raise ValidationError(
                """fragment_separators ({}) unable to perform trial np.split on geometry.""".format(fragment_separators)
            )
        if any(len(f) == 0 for f in split_geom):
            if nat != 0:
                raise ValidationError(
                    """fragment_separators ({}) yields zero-length fragment(s) after trial np.split on geometry.""".format(
                        split_geom
                    )
                )
        if sum(len(f) for f in split_geom) != nat:
            raise ValidationError(
                """fragment_separators ({}) yields overlapping fragment(s) after trial np.split on geometry, possibly unsorted.""".format(
                    split_geom
                )
            )
        frs = fragment_separators
        nfr = len(split_geom)

        if fragment_charges is None:
            frc = [None] * nfr
        else:
            try:
                frc = [(f if f is None else float(f)) for f in fragment_charges]
            except TypeError:
                raise ValidationError("""fragment_charges not among None or float: {}""".format(fragment_charges))

        if fragment_multiplicities is None:
            frm = [None] * nfr
        elif all(f is None or (isinstance(f, (int, np.integer)) and f >= 1) for f in fragment_multiplicities):
            frm = fragment_multiplicities
        else:
            raise ValidationError(
                """fragment_multiplicities not among None or positive integer: {}""".format(fragment_multiplicities)
            )

    if not (len(frc) == len(frm) == len(frs) + 1):
        raise ValidationError(
            """Dimension mismatch among fragment quantities: sep + 1 ({}), chg ({}), and mult({})""".format(
                len(frs) + 1, len(frc), len(frm)
            )
        )

    return {"fragment_separators": list(frs), "fragment_charges": frc, "fragment_multiplicities": frm}


def validate_and_fill_unsettled_geometry(geom_unsettled, variables):
    lgeom = [len(g) for g in geom_unsettled]

    if lgeom[0] not in [0, 3]:
        raise ValidationError("""First line must be Cartesian or single atom.""")

    if any(l == 3 for l in lgeom) and not all((l in [3, 6]) for l in lgeom):
        raise ValidationError(
            """Mixing Cartesian and Zmat formats must occur in just that order once absolute frame established."""
        )

    allowed_to_follow = {0: [2], 2: [4], 3: [3, 6], 4: [6], 6: [3, 6]}

    for il in range(len(lgeom) - 1):
        if lgeom[il + 1] not in allowed_to_follow[lgeom[il]]:
            raise ValidationError(
                """This is not how a Zmat works - aim for lower triangular. Line len ({}) may be followed by line len ({}), not ({}).""".format(
                    lgeom[il], allowed_to_follow[lgeom[il]], lgeom[il + 1]
                )
            )

    if not all(len(v) == 2 for v in variables):
        raise ValidationError("""Variables should come in pairs: {}""".format(variables))

    vvars = [[str(v[0]), float(v[1])] for v in variables]

    return {"geom_unsettled": geom_unsettled, "variables": vvars}



import json
from typing import Any, Union

import numpy as np
from pydantic.json import pydantic_encoder

from .importing import which_import

try:
    import msgpack
except ModuleNotFoundError:
    pass

_msgpack_which_msg = "Please install via `conda install msgpack-python`."

## MSGPackExt


def msgpackext_encode(obj: Any) -> Any:
    """
    Encodes an object using pydantic and NumPy array serialization techniques suitable for msgpack.

    Parameters
    ----------
    obj : Any
        Any object that can be serialized with pydantic and NumPy encoding techniques.

    Returns
    -------
    Any
        A msgpack compatible form of the object.
    """

    # First try pydantic base objects
    try:
        return pydantic_encoder(obj)
    except TypeError:
        pass

    if isinstance(obj, np.ndarray):
        if obj.shape:
            data = {b"_nd_": True, b"dtype": obj.dtype.str, b"data": np.ascontiguousarray(obj).tobytes()}
            if len(obj.shape) > 1:
                data[b"shape"] = obj.shape
            return data

        else:
            # Converts np.array(5) -> 5
            return obj.tolist()

    return obj


def msgpackext_decode(obj: Any) -> Any:
    """
    Decodes a msgpack objects from a dictionary representation.

    Parameters
    ----------
    obj : Any
        An encoded object, likely a dictionary.

    Returns
    -------
    Any
        The decoded form of the object.
    """

    if b"_nd_" in obj:
        arr = np.frombuffer(obj[b"data"], dtype=obj[b"dtype"])
        if b"shape" in obj:
            arr.shape = obj[b"shape"]

        return arr

    return obj


def msgpackext_dumps(data: Any) -> bytes:
    """Safe serialization of a Python object to msgpack binary representation using all known encoders.
    For NumPy, encodes a specialized object format to encode all shape and type data.

    Parameters
    ----------
    data : Any
        A encodable python object.

    Returns
    -------
    bytes
        A msgpack representation of the data in bytes.
    """
    which_import("msgpack", raise_error=True, raise_msg=_msgpack_which_msg)

    return msgpack.dumps(data, default=msgpackext_encode, use_bin_type=True)


def msgpackext_loads(data: bytes) -> Any:
    """Deserializes a msgpack byte representation of known objects into those objects.

    Parameters
    ----------
    data : bytes
        The serialized msgpack byte array.

    Returns
    -------
    Any
        The deserialized Python objects.
    """
    which_import("msgpack", raise_error=True, raise_msg=_msgpack_which_msg)

    return msgpack.loads(data, object_hook=msgpackext_decode, raw=False)


## JSON Ext


class JSONExtArrayEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        try:
            return pydantic_encoder(obj)
        except TypeError:
            pass

        if isinstance(obj, np.ndarray):
            if obj.shape:
                data = {"_nd_": True, "dtype": obj.dtype.str, "data": np.ascontiguousarray(obj).tobytes().hex()}
                if len(obj.shape) > 1:
                    data["shape"] = obj.shape
                return data

            else:
                # Converts np.array(5) -> 5
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def jsonext_decode(obj: Any) -> Any:

    if "_nd_" in obj:
        arr = np.frombuffer(bytes.fromhex(obj["data"]), dtype=obj["dtype"])
        if "shape" in obj:
            arr.shape = obj["shape"]

        return arr

    return obj


def jsonext_dumps(data: Any) -> str:
    """Safe serialization of Python objects to JSON string representation using all known encoders.
    The JSON serializer uses a custom array syntax rather than flat JSON lists.

    Parameters
    ----------
    data : Any
        A encodable python object.

    Returns
    -------
    str
        A JSON representation of the data.
    """

    return json.dumps(data, cls=JSONExtArrayEncoder)


def jsonext_loads(data: Union[str, bytes]) -> Any:
    """Deserializes a json representation of known objects into those objects.

    Parameters
    ----------
    data : str or bytes
        The byte-serialized JSON blob.

    Returns
    -------
    Any
        The deserialized Python objects.
    """

    return json.loads(data, object_hook=jsonext_decode)


## JSON


class JSONArrayEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        try:
            return pydantic_encoder(obj)
        except TypeError:
            pass

        if isinstance(obj, np.ndarray):
            if obj.shape:
                return obj.ravel().tolist()
            else:
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def json_dumps(data: Any) -> str:
    """Safe serialization of a Python dictionary to JSON string representation using all known encoders.

    Parameters
    ----------
    data : Any
        A encodable python object.

    Returns
    -------
    str
        A JSON representation of the data.
    """

    return json.dumps(data, cls=JSONArrayEncoder)


def json_loads(data: str) -> Any:
    """Deserializes a json representation of known objects into those objects.

    Parameters
    ----------
    data : str
        The serialized JSON blob.

    Returns
    -------
    Any
        The deserialized Python objects.
    """

    # Doesn't hurt anything to try to load JSONext as well
    return json.loads(data, object_hook=jsonext_decode)


## Helper functions


def serialize(data: Any, encoding: str) -> Union[str, bytes]:
    """Encoding Python objects using the provided encoder.

    Parameters
    ----------
    data : Any
        A encodable python object.
    encoding : str
        The type of encoding to perform: {'json', 'json-ext', 'msgpack-ext'}

    Returns
    -------
    Union[str, bytes]
        A serialized representation of the data.

    """
    if encoding.lower() == "json":
        return json_dumps(data)
    elif encoding.lower() == "json-ext":
        return jsonext_dumps(data)
    elif encoding.lower() == "msgpack-ext":
        return msgpackext_dumps(data)
    else:
        raise KeyError(f"Encoding '{encoding}' not understood, valid options: 'json', 'json-ext', 'msgpack-ext'")


def deserialize(blob: Union[str, bytes], encoding: str) -> Any:
    """Encoding Python objects using .

    Parameters
    ----------
    blob : Union[str, bytes]
        The serialized data.
    encoding : str
        The type of encoding of the blob: {'json', 'json-ext', 'msgpack'}

    Returns
    -------
    Any
        The deserialized Python objects.
    """
    if encoding.lower() == "json":
        assert isinstance(blob, str)
        return json_loads(blob)
    elif encoding.lower() == "json-ext":
        assert isinstance(blob, (str, bytes))
        return jsonext_loads(blob)
    elif encoding.lower() in ["msgpack", "msgpack-ext"]:
        assert isinstance(blob, bytes)
        return msgpackext_loads(blob)
    else:
        raise KeyError(f"Encoding '{encoding}' not understood, valid options: 'json', 'json-ext', 'msgpack-ext'")



"""
Contains metadata about density functionals
"""

from typing import Dict

from pydantic import Field

from ..models import ProtoModel


class DFTFunctionalInfo(ProtoModel):
    name: str = Field(..., description="The name of the functional.")
    ansatz: int = Field(
        ...,
        description="Maximum functional derivative with respect to the density. This is 2 for meta-GGAs, 1 for GGAs, and 0 otherwise.)",
    )
    deriv: int = Field(..., description="Maximum derivative to compute.")
    c_hybrid: bool = Field(..., description="Requires wavefunction correlation?")
    x_hybrid: bool = Field(..., description="Requires exact exchange?")
    c_lrc: bool = Field(..., description="Contains range-separated correlation?")
    x_lrc: bool = Field(..., description="Contains range-separated exchange?")
    nlc: bool = Field(..., description="Does this functional need non-local correlation?")


class DFTFunctionalContext:
    """Information about DFT functionals.

    Parameters
    ----------
    context : {'default'}
        Origin of loaded data.

    Attributes
    ----------
    functionals : Dict[str, DFTFunctionalInfo]
        dictionary of (lower-case) functional name to info

    name : str
        The name of the context ('default')
    """

    def __init__(self, context: str = "defualt"):
        if context != "default":
            raise KeyError(f"Context set as '{context}'," + " only contexts {'default'} are currently supported")

        from .data import dft_data_blob

        self.suffixes = dft_data_blob.data_blob["empirical_dispersion_suffixes"]
        self.functionals: Dict[str, DFTFunctionalInfo] = {
            name: DFTFunctionalInfo(name=name, **data) for name, data in dft_data_blob.data_blob["functionals"].items()
        }

        self.name = context

    def __str__(self) -> str:
        return "DFTFunctionalContext(context='{}')".format(self.name)


# singleton
dftfunctionalinfo = DFTFunctionalContext("default")


def get(name: str) -> DFTFunctionalInfo:

    name = name.lower()
    for x in dftfunctionalinfo.suffixes:
        if name.endswith(x):
            name = name.replace(x, "")
            break

    return dftfunctionalinfo.functionals[name].copy()



from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import Field, constr, validator

from ..util import provenance_stamp
from .basemodels import ProtoModel
from .common_models import (
    ComputeError,
    DriverEnum,
    Model,
    Provenance,
    qcschema_input_default,
    qcschema_optimization_input_default,
    qcschema_optimization_output_default,
)
from .molecule import Molecule
from .results import AtomicResult

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs


class TrajectoryProtocolEnum(str, Enum):
    """
    Which gradient evaluations to keep in an optimization trajectory.
    """

    all = "all"
    initial_and_final = "initial_and_final"
    final = "final"
    none = "none"


class OptimizationProtocols(ProtoModel):
    """
    Protocols regarding the manipulation of a Optimization output data.
    """

    trajectory: TrajectoryProtocolEnum = Field(
        TrajectoryProtocolEnum.all, description=str(TrajectoryProtocolEnum.__doc__)
    )

    class Config:
        force_skip_defaults = True


class QCInputSpecification(ProtoModel):
    """
    A compute description for energy, gradient, and Hessian computations used in a geometry optimization.
    """

    schema_name: constr(strip_whitespace=True, regex=qcschema_input_default) = qcschema_input_default  # type: ignore
    schema_version: int = 1

    driver: DriverEnum = Field(DriverEnum.gradient, description=str(DriverEnum.__doc__))
    model: Model = Field(..., description=str(Model.__doc__))
    keywords: Dict[str, Any] = Field({}, description="The program specific keywords to be used.")

    extras: Dict[str, Any] = Field({}, description="Extra fields that are not part of the schema.")


class OptimizationInput(ProtoModel):
    id: Optional[str] = None
    hash_index: Optional[str] = None
    schema_name: constr(  # type: ignore
        strip_whitespace=True, regex=qcschema_optimization_input_default
    ) = qcschema_optimization_input_default
    schema_version: int = 1

    keywords: Dict[str, Any] = Field({}, description="The optimization specific keywords to be used.")
    extras: Dict[str, Any] = Field({}, description="Extra fields that are not part of the schema.")
    protocols: OptimizationProtocols = Field(OptimizationProtocols(), description=str(OptimizationProtocols.__doc__))

    input_specification: QCInputSpecification = Field(..., description=str(QCInputSpecification.__doc__))
    initial_molecule: Molecule = Field(..., description="The starting molecule for the geometry optimization.")

    provenance: Provenance = Field(Provenance(**provenance_stamp(__name__)), description=str(Provenance.__doc__))

    def __repr_args__(self) -> "ReprArgs":
        return [
            ("model", self.input_specification.model.dict()),
            ("molecule_hash", self.initial_molecule.get_hash()[:7]),
        ]


class OptimizationResult(OptimizationInput):
    schema_name: constr(  # type: ignore
        strip_whitespace=True, regex=qcschema_optimization_output_default
    ) = qcschema_optimization_output_default

    final_molecule: Optional[Molecule] = Field(..., description="The final molecule of the geometry optimization.")
    trajectory: List[AtomicResult] = Field(
        ..., description="A list of ordered Result objects for each step in the optimization."
    )
    energies: List[float] = Field(..., description="A list of ordered energies for each step in the optimization.")

    stdout: Optional[str] = Field(None, description="The standard output of the program.")
    stderr: Optional[str] = Field(None, description="The standard error of the program.")

    success: bool = Field(
        ..., description="The success of a given programs execution. If False, other fields may be blank."
    )
    error: Optional[ComputeError] = Field(None, description=str(ComputeError.__doc__))
    provenance: Provenance = Field(..., description=str(Provenance.__doc__))

    @validator("trajectory", each_item=False)
    def _trajectory_protocol(cls, v, values):

        # Do not propogate validation errors
        if "protocols" not in values:
            raise ValueError("Protocols was not properly formed.")

        keep_enum = values["protocols"].trajectory
        if keep_enum == "all":
            pass
        elif keep_enum == "initial_and_final":
            if len(v) != 2:
                v = [v[0], v[-1]]
        elif keep_enum == "final":
            if len(v) != 1:
                v = [v[-1]]
        elif keep_enum == "none":
            v = []
        else:
            raise ValueError(f"Protocol `trajectory:{keep_enum}` is not understood.")

        return v


def Optimization(*args, **kwargs):
    from warnings import warn

    warn("Optimization has been renamed to OptimizationResult and will be removed in v0.13.0", DeprecationWarning)
    return OptimizationResult(*args, **kwargs)



"""
Misc information and runtime information.
"""

from . import _version

__all__ = ["get_information"]

versions = _version.get_versions()

__info = {"version": versions["version"], "git_revision": versions["full-revisionid"]}


def get_information(key: str):
    """
    Obtains a variety of runtime information about QCElemental.
    """
    key = key.lower()
    if key not in __info:
        raise KeyError(f"Information key '{key}' not understood.")

    return __info[key]
